"""Models related to plans, plan associations, and wires."""
import json
from warnings import warn

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import DeleteMixin
from pydent.models.crud_mixin import SaveMixin
from pydent.models.data_associations import DataAssociatorMixin
from pydent.models.data_associations import Upload
from pydent.models.field_value import FieldValue
from pydent.relationships import HasMany
from pydent.relationships import HasManyGeneric
from pydent.relationships import HasManyThrough
from pydent.relationships import HasOne
from pydent.relationships import JSON
from pydent.relationships import Many
from pydent.relationships import Raw


@add_schema
class Plan(DataAssociatorMixin, SaveMixin, DeleteMixin, ModelBase):
    """A Plan model."""

    fields = dict(
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Plan"}
        ),
        plan_associations=HasMany("PlanAssociation", "Plan"),
        operations=HasManyThrough("Operation", "PlanAssociation"),
        wires=Many("Wire", callback="_get_wires_from_server"),
        layout=JSON(),
        status=Raw(default="planning"),
        user=HasOne("User"),
    )
    query_hook = {"include": ["plan_associations", "operations"]}

    def __init__(self, name="MyPlan", status="planning"):
        super().__init__(
            name=name,
            status=status,
            data_associations=None,
            plan_associations=None,
            layout={
                "id": 0,
                "children": [],
                "documentation": "No documentation of this module",
                "height": 60,
                "input": [],
                "output": [],
                "name": "no_name",
                "parent_id": -1,
                "width": 160,
                "wires": [],
            },
        )

    def add_operation(self, operation):
        """Adds an operation to the Plan.

        :param operation: Operation to add
        :type operation: Operation
        :return: None
        :rtype: None
        """
        self.append_to_many("operations", operation)

    def add_operations(self, operations):
        """Adds multiple operations to the Plan.

        :param operations: list of Operations
        :type operations: list
        :return: None
        :rtype: None
        """
        for operation in operations:
            self.add_operation(operation)

    # TODO: this is not functional or not needed
    # def has_operation(self, op):
    #     if op is None:
    #         return False
    #     return self.operations and op.rid in [_op.rid for _op in self.operations]

    def find_wires(self, src, dest):
        """Retrieves the wire between a source and destination FieldValues.

        :param src: source FieldValue
        :type src: FieldValue
        :param dest: destination FieldValue
        :type dest: FieldValue
        :return: array of wires between src and dest FieldValues (determined by rid)
        :rtype: array of wires
        """

        found = []
        for wire in self.wires:
            if src.rid == wire.source.rid and dest.rid == wire.destination.rid:
                found.append(wire)
        return found

    def wire(self, src, dest):
        """Creates a new wire between src and dest FieldValues. Returns the new
        wire if it does not exist in the plan. If the wire already exists and
        error_if_exists is True, then the existing wire is returned. Else an
        exception is raised.

        :param src: source field value (the input of the wire)
        :type src: FieldValue
        :param dest: destination field value (the output of the wire)
        :type dest: FieldValue
        :param error_if_exists: Raise an error if the Wire already exists in the plan.
        :type error_if_exists: boolean
        :return: Newly created wire or existing wire (if exists and
            error_if_exists == False)
        :rtype: Wire
        """

        # TODO: these checks are unnecessary?
        # if not self.has_operation(src.operation):
        #     raise AquariumModelError(
        #         "Cannot wire because the wire's source FieldValue {} does "
        #         "not exist in the Plan because its Operation '{}' is not in the plan".format(
        #             src, src.operation
        #         )
        #     )
        # if not self.has_operation(dest.operation):
        #     raise AquariumModelError(
        #         "Cannot wire because the wire's destination FieldValue {} does not "
        #         "exist in the Plan because its Operation '{}' is not in the plan.".format(
        #             dest, dest.operation
        #         )
        #     )

        wire = Wire(source=src, destination=dest)
        self.append_to_many("wires", wire)
        return wire

    def _collect_wires(self):
        incoming_wires = []
        outgoing_wires = []
        if self.operations:
            for operation in self.operations:
                if operation.field_values:
                    for field_value in operation.field_values:
                        if field_value.outgoing_wires:
                            outgoing_wires += field_value.outgoing_wires
                        if field_value.incoming_wires:
                            incoming_wires += field_value.incoming_wires
        return incoming_wires, outgoing_wires

    def _get_wire_dict(self, wires):
        """Return all wires in the plan grouped by the wire identifier."""
        wire_dict = {}
        for w in wires:
            wire_dict.setdefault(w.identifier, list())
            if w not in wire_dict[w.identifier]:
                wire_dict[w.identifier].append(w)
        return wire_dict

    def _get_wires_from_server(self, *args):
        fvs = []
        if self.operations:
            for op in self.operations:
                fvs += op.field_values

        fv_ids = [fv.id for fv in fvs if fv.id is not None]
        wires_from_server = []
        if fv_ids:
            wires_from_server = self.session.Wire.where(
                {"from_id": fv_ids, "to_id": fv_ids}
            )
        return wires_from_server

    def submit(self, user, budget):
        """Submits the Plan to the Aquarium server.

        :param user: User to submit the Plan
        :type user: User
        :param budget: Budget to use for the Plan
        :type budget: Budget
        :return: JSON
        :rtype: dict
        """
        result = self.session.utils.submit_plan(self, user, budget)
        return result

    def all_data_associations(self):
        das = self.data_associations
        for operation in self.operations:
            das += operation.data_associations
            for field_value in operation.field_values:
                if field_value.item:
                    das += field_value.item.data_associations
        return das

    @classmethod
    def interface(cls, session):
        # get model interface from Base class
        model_interface = super().interface(session)

        # make a special find method for plans, as generic method is too minimal.
        def new_find(model_id):
            return model_interface.get("plans/{}.json".format(model_id))

        # override the old find method
        model_interface.find = new_find

        return model_interface

    def validate(self, raise_error=True):
        """Validates the plan.

        :param raise_error: If True, raises an AquariumModelException. If false,
            returns the error messages.
        :type raise_error: boolean
        :return: list of error messages
        :rtype: array
        """
        errors = []

        field_values = []
        for op in self.operations:
            for fv in op.field_values:
                field_values.append(fv)
        fv_keys = [fv._primary_key for fv in field_values]
        for wire in self.wires:
            for _fvtype in ["source", "destination"]:
                field_value = getattr(wire, _fvtype)
                if field_value._primary_key not in fv_keys:
                    msg = (
                        "The FieldValue of a wire Wire(rid={}).{} is missing from the "
                        "list of FieldValues in the plan. Did you forget to add an "
                        "operation to the plan?".format(wire._primary_key, _fvtype)
                    )
                    errors.append(msg)
        if raise_error and errors:
            msg = "\n".join(
                ["(ErrNo {}) - {}".format(i, e) for i, e in enumerate(errors)]
            )
            raise AquariumModelError(
                "Plan {} had the following errors:\n{}".format(self, msg)
            )
        return errors

    def to_save_json(self):
        """Returns the json representation of the plan for saving and creating
        Plans on the Aquarium server.

        :return: JSON
        :rtype: dict
        """
        if not self.operations:
            self.operations = []

        if not self.wires:
            self.wires = []

        self.validate(raise_error=True)

        for op in self.operations:
            op.field_values

        json_data = self.dump(
            include={"operations": {"field_values": ["sample", "item"]}}
        )

        # remove redundant wires
        wire_dict = {}
        for wire in self.wires:
            wire_data = wire.to_save_json()
            attributes = [
                wire_data["from_id"],
                wire_data["from"]["rid"],
                wire_data["to_id"],
                wire_data["to"]["rid"],
            ]
            wire_hash = "*&".join([str(a) for a in attributes])
            wire_dict[wire_hash] = wire_data
        json_data["wires"] = list(wire_dict.values())

        # validate
        fv_rids = []
        fv_id_to_rids = {}
        for op in json_data["operations"]:
            for fv in op["field_values"]:
                if fv["id"]:
                    fv_id_to_rids[fv["id"]] = fv["rid"]
                fv_rids.append(fv["rid"])

        # fix json rids and ids
        warnings = []
        for wire_data in json_data["wires"]:
            if wire_data["from_id"] in fv_id_to_rids:
                wire_data["from"]["rid"] = fv_id_to_rids[wire_data["from_id"]]
            if wire_data["to_id"] in fv_id_to_rids:
                wire_data["to"]["rid"] = fv_id_to_rids[wire_data["to_id"]]
            if not wire_data["from"]["rid"] in fv_rids:
                warnings.append(
                    "FieldValue rid={} is missing.".format(wire_data["from"]["rid"])
                )
            if not wire_data["to"]["rid"] in fv_rids:
                warnings.append(
                    "FieldValue rid={} is missing.".format(wire_data["to"]["rid"])
                )
        if warnings:
            print(fv_rids)
            warn(",".join(warnings))

        if json_data["layout"] is not None:
            json_data["layout"] = json.loads(json_data["layout"])
        else:
            del json_data["layout"]

        return json_data

    def _get_create_json(self):
        return self.to_save_json()

    def _get_update_json(self):
        return self.to_save_json()

    def _get_create_params(self):
        return {"user_id": self.session.current_user.id}

    def _get_update_params(self):
        return {"user_id": self.session.current_user.id}

    def estimate_cost(self):
        """Estimates the cost of the plan on the Aquarium server. This is
        necessary before plan submission.

        :return: cost
        :rtype: dict
        """
        return self.session.utils.estimate_plan_cost(self)

    def field_values(self):
        raise NotImplementedError()

    def step(self):
        """Steps a plan."""
        return self.session.utils.step_plan(self.id)

    def show(self):
        """Print the plan nicely."""
        print(self.name + " id: " + str(self.id))
        for operation in self.operations:
            operation.show(pre="  ")
        for wire in self.wires:
            wire.show(pre="  ")

    def replan(self):
        """Copies or replans the plan.

        Returns a plan copy
        """
        return self.session.utils.replan(self.id)

    def download_files(self, outdir=None, overwrite=True):
        """Downloads all uploads associated with the plan. Downloads happen
        ansynchrounously.

        :param outdir: output directory for downloaded files
        :param overwrite: whether to overwrite files if they exist
        :return: None
        """
        uploads = []
        for da in self.data_associations:
            if da.upload is not None:
                uploads.append(da.upload)
        return Upload.async_download(uploads, outdir, overwrite)


@add_schema
class PlanAssociation(ModelBase):
    """A PlanAssociation model."""

    fields = dict(plan=HasOne("Plan"), operation=HasOne("Operation"))

    def __init__(self, plan_id=None, operation_id=None):
        super().__init__(plan_id=plan_id, operation_id=operation_id)


@add_schema
class Wire(DeleteMixin, ModelBase):
    """A Wire model."""

    fields = {
        "source": HasOne("FieldValue", ref="from_id"),
        "destination": HasOne("FieldValue", ref="to_id"),
    }

    WIRABLE_PARENT_CLASSES = ["Operation"]

    def __init__(self, source=None, destination=None):

        self._validate_field_values(source, destination)

        if hasattr(source, "id"):
            from_id = source.id
        else:
            from_id = None

        if hasattr(destination, "id"):
            to_id = destination.id
        else:
            to_id = None

        if (source and not destination) or (destination and not source):
            raise AquariumModelError(
                "Cannot wire. Either source ({}) or destination ({}) is None".format(
                    source, destination
                )
            )

        if source:
            if not source.role:
                raise AquariumModelError(
                    "Cannot wire. FieldValue {} does not have a role".format(
                        source.role
                    )
                )
            elif source.role != "output":
                raise AquariumModelError(
                    "Cannot wire an '{}' FieldValue as a source".format(source.role)
                )

        if destination:
            if not destination.role:
                raise AquariumModelError(
                    "Cannot wire. FieldValue {} does not have a role".format(
                        destination.role
                    )
                )
            elif destination.role != "input":
                raise AquariumModelError(
                    "Cannot wire an '{}' FieldValue as a destination".format(
                        destination.role
                    )
                )

        super().__init__(
            **{
                "source": source,
                "from_id": from_id,
                "destination": destination,
                "to_id": to_id,
            },
            active=True,
        )

    @property
    def identifier(self):

        if not self.source:
            source_id = self.from_id
        else:
            source_id = "r" + str(self.source.rid)

        if not self.destination:
            destination_id = self.to_id
        else:
            destination_id = "r" + str(self.destination.rid)
        return "{}_{}".format(source_id, destination_id)

    @classmethod
    def _validate_field_values(cls, src, dest):
        if src:
            cls._validate_field_value(src, "source")
        if dest:
            cls._validate_field_value(dest, "destination")

        if src and dest and src.rid == dest.rid:
            raise AquariumModelError(
                "Cannot create wire because source and destination are the same "
                "instance."
            )

    @classmethod
    def _validate_field_value(cls, fv, name):
        if not issubclass(type(fv), FieldValue):
            raise AquariumModelError(
                "Cannot create wire because {} FieldValue is {}.".format(
                    name, fv.__class__.__name__
                )
            )

        if fv.parent_class not in cls.WIRABLE_PARENT_CLASSES:
            raise AquariumModelError(
                "Cannot create wire because the {} FieldValue is has '{}' "
                "parent class. Only {} parent classes are wirable".format(
                    name, fv.parent_class, cls.WIRABLE_PARENT_CLASSES
                )
            )

    def validate(self):
        self._validate_field_values(self.source, self.destination)

    def to_save_json(self):
        save_json = {
            "id": self.id,
            "from_id": self.source.id,
            "to_id": self.destination.id,
            "from": {"rid": self.source.rid},
            "to": {"rid": self.destination.rid},
            "active": self.active,
        }
        return save_json

    def delete(self):
        """Permanently deletes the wire instance on the Aquarium server.

        :return:
        :rtype:
        """
        return self.session.utils.delete_wire(self)

    def show(self, pre=""):
        """Show the wire nicely."""
        source = self.source
        dest = self.destination

        from_op_type_name = "None"
        from_name = "None"
        if source:
            from_op_type_name = self.source.operation.operation_type.name
            from_name = source.name

        to_op_type_name = "None"
        to_name = "None"
        if dest:
            to_op_type_name = self.destination.operation.operation_type.name
            to_name = dest.name

        print(
            pre
            + from_op_type_name
            + ":"
            + from_name
            + " --> "
            + to_op_type_name
            + ":"
            + to_name
        )

    def does_wire_to(self, destination):
        if destination and self.destination:
            return destination.rid == self.destination.rid
        return False

    def does_wire_from(self, source):
        if source and self.source:
            return source.rid == self.source.rid
        return False

    def does_wire(self, source, destination):
        """Checks whether this Wire is a wire between the source and
        destination FieldValues.

        If any of the source or destination FieldValues are None,
        returns False.
        """
        if source and destination and self.source and self.destination:
            return (
                source.rid == self.source.rid
                and destination.rid == self.destination.rid
            )
        return False

    # def __eq__(self, other):
    #     """Checks whether this Wire is wired to the same FieldValue instances
    #     of another Wire.
    #
    #     see `does_wire` method.
    #     """
    #     return self.does_wire(other.source, other.destination)



# __version__.py

# autogenerated by keats 0.2.29

__version__ = "1.0.9"
__name__ = "pydent"
__title__ = "pydent"
__authors__ = ['jvrana <justin.vrana@gmail.com>', 'Ben Keller <bjkeller@uw.edu>', 'Eric Klavins <klavins@uw.edu>']
__repository__ = "https://www.github.com/aquariumbio/trident"
__homepage__ = "https://www.github.com/aquariumbio/trident"
__description__ = "Aquarium's Python API for planning, executing, and analyzing scientific experiments."
__license__ = "MIT"
__maintainers__ = None
__readme__ = "README.md"


import difflib
from pprint import pformat

from pydent.utils import condense_long_lists


def did_you_mean(word, words):
    msg = ""
    try:
        matches = difflib.get_close_matches(word, words)
        if len(matches) == 1:
            msg = 'Did you mean "{}"?'.format(matches[0])
        elif len(matches) > 1:
            msg = "Did you mean any of these? {}".format(pformat(matches))
    except TypeError:
        pass
    finally:
        return msg



"""Models related to operations and operation types."""
from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import SaveMixin
from pydent.models.data_associations import DataAssociatorMixin
from pydent.models.field_value_mixins import FieldTypeInterface
from pydent.models.field_value_mixins import FieldValueInterface
from pydent.relationships import Function
from pydent.relationships import HasMany
from pydent.relationships import HasManyGeneric
from pydent.relationships import HasManyThrough
from pydent.relationships import HasOne
from pydent.relationships import HasOneFromMany
from pydent.relationships import Raw
from pydent.utils import filter_list


@add_schema
class Operation(FieldValueInterface, DataAssociatorMixin, ModelBase):
    """A Operation model."""

    fields = dict(
        field_values=HasMany(
            "FieldValue", ref="parent_id", additional_args={"parent_class": "Operation"}
        ),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Operation"}
        ),
        operation_type=HasOne("OperationType"),
        job_associations=HasMany("JobAssociation", "Operation"),
        jobs=HasManyThrough("Job", "JobAssociation"),
        plan_associations=HasMany("PlanAssociation", "Operation"),
        plans=HasManyThrough("Plan", "PlanAssociation"),
        status=Raw(default="planning"),
        routing=Function("get_routing"),
        user=HasOne("User"),
    )

    METATYPE = "operation_type"

    def __init__(
        self, operation_type_id=None, operation_type=None, status=None, x=0, y=0
    ):
        super().__init__(
            operation_type_id=operation_type_id,
            operation_type=operation_type,
            status=status,
            field_values=None,
            x=x,
            y=y,
        )

    def get_routing(self):
        routing_dict = {}
        fvs = self.field_values
        ot = self.operation_type
        if ot is None:
            return routing_dict
        for ft in ot.field_types:
            if ft.routing is not None:
                routing_dict[ft.routing] = None
        if fvs is not None:
            for fv in self.field_values:
                ft = self.safe_get_field_type(fv)
                if ft.routing is not None:
                    routing_dict[ft.routing] = fv.sid
        return routing_dict

    @property
    def successors(self):
        successors = []
        if self.outputs:
            for output in self.outputs:
                for s in output.successors:
                    successors.append(s.operation)
        return successors

    @property
    def predecessors(self):
        predecessors = []
        if self.inputs:
            for inputs in self.inputs:
                for s in inputs.predecessors:
                    predecessors.append(s.operation)
        return predecessors

    def init_field_values(self):
        """Initialize the :class:`FieldValue` from the :class:`FieldType` of
        the parent :class:`Operation` type."""
        self.field_values = []
        for field_type in self.get_metatype().field_types:
            if not field_type.array:
                self.new_field_value_from_field_type(field_type)
        return self

    def field_value_array(self, name, role):
        """Returns :class:`FieldValue` array with name and role."""
        return filter_list(self.get_field_value_array(name, role))

    def field_value(self, name, role):
        """Returns :class:`FieldValue` with name and role.

        Return None if not found.
        """
        if self.field_values:
            fvs = self.field_value_array(name, role)

            if len(fvs) == 0:
                return None

            if len(fvs) == 1:
                return fvs[0]

            msg = "More than one FieldValue found for the field value"
            msg += (
                " of operation {}.(id={}).{}.{}. Are you sure you didn't mean to "
                "call 'field_value_array'?"
            )
            raise AquariumModelError(
                msg.format(self.operation_type, self.id, role, name)
            )

    @property
    def plan(self):
        return self.plans[0]

    def input_array(self, name):
        return self.get_field_value_array(name, "input")

    def output_array(self, name):
        return self.get_field_value_array(name, "output")

    def input(self, name):
        """Returns the input :class:`FieldValue` by name."""
        return self.field_value(name, "input")

    def output(self, name):
        """Returns the output :class:`FieldValue` by name."""
        return self.field_value(name, "output")

    def add_to_input_array(
        self, name, sample=None, item=None, value=None, container=None
    ):
        """Creates and adds a new input :class:`FieldValue`. When setting
        values to items/samples/containers, the item/sample/container must be
        saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.new_field_value(
            name,
            "input",
            dict(sample=sample, item=item, value=value, container=container),
        )

    def add_to_output_array(
        self, name, sample=None, item=None, value=None, container=None
    ):
        """Creates and adds a new output :class:`FieldValue`. When setting
        values to items/samples/containers, the item/sample/container must be
        saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.new_field_value(
            name,
            "output",
            dict(sample=sample, item=item, value=value, container=container),
        )

    @property
    def inputs(self):
        """Return a list of all input :class:`FieldValues`"""
        return [fv for fv in self.field_values if fv.role == "input"]

    @property
    def outputs(self):
        """Return a list of all output :class:`FieldValues`"""
        return [fv for fv in self.field_values if fv.role == "output"]

    def set_input(
        self, name, sample=None, item=None, value=None, container=None, object_type=None
    ):
        """Sets a input :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        if object_type is None and container:
            object_type = container
        return self.set_field_value(
            name,
            "input",
            dict(sample=sample, item=item, value=value, object_type=container),
        )

    def set_output(
        self, name, sample=None, item=None, value=None, container=None, object_type=None
    ):
        """Sets a output :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        if object_type is None and container:
            object_type = container
        return self.set_field_value(
            name,
            "output",
            dict(sample=sample, item=item, value=value, object_type=object_type),
        )

    def set_input_array(self, name, values):
        """Sets input :class:`FieldValue` array using values. Values should be
        a list of dictionaries containing sample, item, container, or values
        keys. When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "input", values)

    def set_output_array(self, name, values):
        """Sets output :class:`FieldValue` array using values. Values should be
        a list of dictionaries containing sample, item, container, or values
        keys. When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "output", values)

    def show(self, pre=""):
        """Print the operation nicely."""
        print(pre + self.operation_type.name + " " + str(self.cost))
        for field_value in self.field_values:
            field_value.show(pre=pre + "  ")

    def __str__(self):
        return self._to_str(operation_type_name=self.operation_type.name)


@add_schema
class OperationType(FieldTypeInterface, SaveMixin, ModelBase):
    """Represents an OperationType, which is the definition of a protocol in
    Aquarium."""

    fields = dict(
        operations=HasMany("Operation", "OperationType"),
        field_types=HasMany(
            "FieldType",
            ref="parent_id",
            additional_args={"parent_class": "OperationType"},
        ),
        codes=HasManyGeneric("Code"),
        cost_model=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "cost_model"},
        ),
        documentation=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "documentation"},
        ),
        precondition=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "precondition"},
        ),
        protocol=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "protocol"},
        ),
        test=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "test"},
        ),
    )

    def code(self, accessor):
        if accessor in [
            "protocol",
            "precondition",
            "documentation",
            "cost_model",
            "test",
        ]:
            return getattr(self, accessor)
        return None

    def instance(self, xpos=0, ypos=0):
        operation = self.session.Operation.new(
            operation_type_id=self.id, status="planning", x=xpos, y=ypos
        )
        operation.operation_type = self
        operation.init_field_values()
        return operation

    def field_type(self, name, role):
        if self.field_types:
            fts = filter_list(self.field_types, role=role, name=name)
            if len(fts) > 0:
                return fts[0]

    def sample_type(self):
        sample_types = []
        for field_type in self.field_types:
            for allowable_field_type in field_type.allowable_field_types:
                sample_types.append(allowable_field_type.sample_type)
        return sample_types

    def object_type(self):
        object_types = []
        for field_type in self.field_types:
            for allowable_field_type in field_type.allowable_field_types:
                object_types.append(allowable_field_type.object_type)
        return object_types

    def output(self, name):
        return self.field_type(name, "output")

    def input(self, name):
        return self.field_type(name, "input")

    def save(self):
        """Saves the Operation Type to the Aquarium server.

        Requires this Operation Type to be connected to a session.
        """
        return self.reload(self.session.utils.create_operation_type(self))

    def to_save_json(self):
        op_data = self.dump(
            include={
                "field_types": {"allowable_field_types": {}},
                "protocol": {},
                "cost_model": {},
                "documentation": {},
                "precondition": {},
                "test": {},
            }
        )

        # Format 'sample_type' and 'object_type' keys for afts
        for ft_d, ft in zip(op_data["field_types"], self.field_types):
            for aft_d, aft in zip(
                ft_d["allowable_field_types"], ft.allowable_field_types
            ):
                aft_d["sample_type"] = {"name": aft.sample_type.name}
                aft_d["object_type"] = {"name": aft.object_type.name}
        return op_data

    def _get_create_json(self):
        return self.to_save_json()

    def _get_update_json(self):
        return self.to_save_json()

    def __str__(self):
        return self._to_str("id", "name", "category")



class QueryBuilder:
    class Not:
        def __init__(self, v):
            self.v = v

    @classmethod
    def sql(cls, data):
        rows = []
        for k, v in data.items():
            if not isinstance(v, cls.Not):
                rows.append('{} = "{}"'.format(k, v))
            else:
                rows.append('{} != "{}"'.format(k, v.v))
        return " AND ".join(rows)



"""Mixins for FieldValues and those objects using FieldValues and
FieldTypes."""
from collections.abc import Sequence
from itertools import zip_longest

from pydent.base import ModelBase
from pydent.utils import filter_list


class FieldMixin:
    """Mixin for finding FieldType and FieldValue relationships."""

    def find_field_parent(self, model_name, model_id):
        """Callback for finding operation_type or sample_type.

        If parent_class does not match the expected nested model name
        (OperationType or SampleType), callback will return None
        """
        if model_name == self.parent_class:
            # weird, but helps with refactoring this mixin
            fxn_name = ModelBase.find_callback.__name__
            fxn = getattr(self, fxn_name)
            return fxn(model_name, model_id)


class FieldTypeInterface:
    """An interface for things that have field types (i.e.

    :class:`pydent.models.OperationType` and :class:`pydent.models.SampleType`).
    """

    def field_type(self, name, role):
        """Returns a :class:`pydent.models.FieldType` by its name and role.

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :return: the field type or None if not found
        :rtype: FieldType | None
        """
        if self.field_types:
            fts = filter_list(self.field_types, role=role, name=name)
            if len(fts) > 0:
                return fts[0]


class FieldValueInterface:
    """A common interface for things (i.e. Operations and Samples) that have
    FieldValues and FieldTypes."""

    def new_field_value_from_field_type(self, field_type, values_dict=None):
        """Instantiates a new :class:`pydent.models.FieldValue` from a.

        :class:`pydent.models.FieldType`
        instance. Optionally, a values dictionary may be provided with the following
        format:

        ::

            values_dict = {
                "value": None,
                "sample": mysample,
                "item": myitem,
                "object_type": None
            }

        :param field_type: field type to instantiate the field value
        :type field_type: FieldType
        :param values_dict: values to set the new field value
        :type values_dict: dict
        :return: the new field value
        :rtype: FieldValue
        """

        assert field_type in self.get_field_types()
        self.new_field_value(field_type.name, field_type.role, values_dict=values_dict)

    def new_field_value(self, name, role=None, values_dict=None):
        """
        Instantiates a new :class:`pydent.models.FieldValue` from a name and role.
        The should be a field type with the name and role
        in the metatype.
        Optionally, a values dictionary may be provided with the following format:

        ::

            values_dict = {
                "value": None,
                "sample": mysample,
                "item": myitem,
                "object_type": None
            }

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :param values_dict: values to set the new field value
        :type values_dict: dict
        :return: the new field value
        :rtype: FieldValue
        """

        # retrieve the field_type from the meta_type
        metatype = self.get_metatype()
        ft = metatype.field_type(name, role=role)

        # initialize the field_value
        fv = ft.initialize_field_value(parent=self)
        if values_dict:
            fv.set_value(**values_dict)
        # add new field_value to list of field_values
        if self.field_values is None:
            self.field_values = []

        self.field_values.append(fv)
        return fv

    def safe_get_field_type(self, fv):
        """Safely returns the field value's :class:`pydent.models.FieldType`
        from the model.

        If the field value has no reference to the field type, its
        metatype (e.g. :class:`pydent.models.SampleType` or
        :class:`pydent.models.OperationType`) is used to recover the
        field type.
        """

        def h(f):
            return "{}_%&^_{}".format(f.name, f.role)

        if fv.field_type_id is None:
            fts = self.get_field_types()
            name_role_to_ft = {h(ft): ft for ft in fts}
            ft = name_role_to_ft.get(h(fv), None)
            if ft is not None:
                fv.set_field_type(ft)
        return fv.field_type

    def get_metatype(self):
        """Returns the instance's metatype.

        The metatype is either
        :class:`pydent.models.SampleType` or
        :class:`pydent.models.OperationType`). The metatype class name is
        stored as `METATYPE` in the class definition.

        :return: the metatype
        :rtype: ModelBase
        """

        metatype = getattr(self, self.METATYPE)
        assert issubclass(type(metatype), FieldTypeInterface)
        return metatype

    def get_field_value(self, name, role=None):
        """Returns a :class:`pydent.models.FieldValue` by its name and role.

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :return: the field value or None if not found
        :rtype: FieldValue | None
        """

        fv_array = self.get_field_value_array(name, role=role)
        if fv_array:
            return fv_array[0]

    def get_field_value_array(self, name, role=None):
        """Returns a list of :class:`pydent.models.FieldValue` by their name
        and role.

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :return: list of field values
        :rtype: list
        """

        if self.field_values is None:
            return []
        fvs = []
        for fv in self.field_values:
            self.safe_get_field_type(fv)
            if fv.name == name and fv.role == role:
                self.safe_get_field_type(fv)
                fvs.append(fv)
        return fvs

    def get_field_types(self, name: str = None, role: str = None):
        """Returns a list of :class:`pydent.models.FieldType` by their name and
        role from the instance's metaclass.

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :return: list of field types
        :rtype: list
        """

        fts = self.get_metatype().field_types
        if name is not None:
            fts = [ft for ft in fts if ft.name == name]
        if role is not None:
            fts = [ft for ft in fts if ft.role == role]
        return fts

    def get_field_type(self, name, role):
        """Returns a :class:`pydent.models.FieldType` by its name and role from
        the instance's metaclass.

        :param name: its name
        :type name: basestring
        :param role: its role
        :type role: basestring
        :return: metatypes field type or None if not found
        :rtype: FieldType | None
        """

        fts = self.get_field_types(name, role)
        if fts:
            return fts[0]

    def _field_value_dictionary(self, ft_func=None, fv_func=None):

        if ft_func is None:

            def ft_func(ft):
                return ft.name

        if fv_func is None:

            def fv_func(fv):
                return fv

        data = {}
        ft_dict = {ft.id: ft for ft in self.get_field_types()}

        for ft in ft_dict.values():
            if ft.array:
                val = []
            else:
                val = None
            data[ft_func(ft)] = val

        # safely get field_types; field_type_id is sometimes None in Sample field_
        # values; unsure if this used to be a bug in Aquarium that has been since
        # resolved

        fvs = self.field_values
        if fvs:
            for fv in fvs:
                val = fv_func(fv)
                ft = self.safe_get_field_type(fv)
                if ft:
                    if ft.array:
                        data[ft_func(ft)].append(val)
                    else:
                        data[ft_func(ft)] = val

        return data

    def set_field_value(self, name, role, values):
        """
        Sets a field value name and role using a dictionary, as in the following:

        ::

            values = {
                "value": None,
                "sample": mysample,
                "item": myitem,
                "object_type": None
            }
        """

        ft = self.get_field_type(name, role)
        fv = self.get_field_value(name, role)
        if fv is None:
            self.new_field_value_from_field_type(ft, values)
        else:
            fv.set_value(**values)
        return self

    def set_field_value_array(self, name, role, values_array):
        """
        Sets an array of field values by name and role using an array of dictionaries,
        as in the following:

        ::

            values_array = [{
                "value": None,
                "sample": mysample,
                "item": myitem,
                "object_type": None
            }]
        """

        fvs = self.get_field_value_array(name, role)
        to_be_removed = []
        for fv, val in zip_longest(fvs, values_array):
            if fv and val:
                fv.set_value(**val)
            elif fv and not val:
                to_be_removed.append(fv)
            elif not fv and val:
                self.new_field_value(name, role, val)
        for fv in to_be_removed:
            self.field_values.remove(fv)
        return self

    def get_routing(self):
        """Returns the routing dictionary for this instance.

        :return: routing dictionary
        :rtype: dict
        """
        return self._field_value_dictionary(lambda ft: ft.routing, lambda fv: fv.sid)

    def update_field_values(self, value_dict, role=None):
        ft_dict = {ft.name: ft for ft in self.get_field_types()}

        for name, val in value_dict.items():
            ft = ft_dict[name]
            if issubclass(type(val), Sequence) and ft.array:
                self.set_field_value_array(name, role, val)
            else:
                self.set_field_value(name, role, val)
        return self



"""Models related to field values and sample properties."""
import json
from warnings import warn

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.exceptions import TridentDepreciationWarning
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import JSONDeleteMixin
from pydent.models.crud_mixin import JSONSaveMixin
from pydent.models.field_value_mixins import FieldMixin
from pydent.models.inventory import Collection
from pydent.models.inventory import Item
from pydent.models.inventory import ObjectType
from pydent.models.sample import Sample
from pydent.relationships import fields
from pydent.relationships import Function
from pydent.relationships import HasMany
from pydent.relationships import HasOne
from pydent.relationships import Raw
from pydent.utils import filter_list


class NullType:
    class __NullType:
        def __bool__(self):
            return False

    instance = None

    def __new__(cls):
        if NullType.instance is None:
            NullType.instance = NullType.__NullType()
        return NullType.instance

    def __bool__(self):
        return False


Null = NullType()


@add_schema
class AllowableFieldType(ModelBase):
    """A AllowableFieldType model."""

    fields = dict(
        field_type=HasOne("FieldType"),
        object_type=HasOne("ObjectType"),
        sample_type=HasOne("SampleType"),
    )

    def __init__(self, field_type=None, object_type=None, sample_type=None):
        super().__init__(
            field_type_id=None,
            sample_type_id=None,
            object_type_id=None,
            field_type=field_type,
            object_type=object_type,
            sample_type=sample_type,
        )

    def __str__(self):
        return self._to_str("sample_type", "object_type")


@add_schema
class FieldType(FieldMixin, ModelBase):
    """A FieldType model."""

    fields = dict(
        allowable_field_types=HasMany("AllowableFieldType", "FieldType"),
        operation_type=HasOne(
            "OperationType", callback="find_field_parent", ref="parent_id"
        ),
        sample_type=HasOne("SampleType", callback="find_field_parent", ref="parent_id"),
        field_values=HasMany("FieldValue", "FieldType"),
    )

    def __init__(
        self,
        name=None,
        ftype=None,
        array=None,
        choices=None,
        operation_type=None,
        preferred_field_type_id=None,
        preferred_operation_type_id=None,
        required=None,
        routing=None,
        role=None,
        parent_class=None,
        parent_id=None,
        sample_type=None,
        aft_stype_and_objtype=(),
        allowable_field_types=None,
    ):
        if operation_type and sample_type:
            raise AquariumModelError(
                "Cannot instantiate a FieldType for both a OperationType and SampleType."
            )
        if operation_type:
            parent_class = "OperationType"
        if sample_type:
            parent_class = "SampleType"
        super().__init__(
            name=name,
            ftype=ftype,
            array=array,
            choices=choices,
            role=role,
            preferred_field_type_id=preferred_field_type_id,
            preferred_operation_type_id=preferred_operation_type_id,
            required=required,
            routing=routing,
            parent_class=parent_class,
            parent_id=parent_id,
            sample_type=sample_type,
            operation_type=operation_type,
            allowable_field_types=allowable_field_types,
        )
        self.part = None
        if allowable_field_types is None:
            if aft_stype_and_objtype is not None:
                for smple_type, obj_type in aft_stype_and_objtype:
                    self.create_allowable_field_type(smple_type, obj_type)

    def get_choices(self):
        if self.choices == "":
            return None
        if self.choices is not None:
            return self.choices.split(",")

    def is_parameter(self):
        return self.ftype != "sample"

    def get_allowable_field_types(self, sample_type_id=None, object_type_id=None):
        afts = self.allowable_field_types
        if afts is None or afts == []:
            self.allowable_field_types = None
            afts = self.allowable_field_types
        return filter_list(
            afts, sample_type_id=sample_type_id, object_type_id=object_type_id
        )

    def create_allowable_field_type(self, sample_type=None, object_type=None):
        afts = []
        if self.allowable_field_types:
            afts = self.allowable_field_types
        field_type = self.session.AllowableFieldType.new(
            field_type=self, sample_type=sample_type, object_type=object_type
        )
        afts.append(field_type)
        self.allowable_field_types = afts
        return field_type

    def initialize_field_value(self, field_value=None, parent=None):
        """Updates or initializes a new :class:`FieldValue` from this
        FieldType.

        :param field_value: optional FieldValue to update with this FieldType
        :type field_value: FieldValue
        :return: updated FieldValue
        :rtype: FieldValue
        """

        if not field_value:
            field_value = self.session.FieldValue.new(
                name=self.name, role=self.role, field_type=self
            )
        if self.allowable_field_types:
            field_value.allowable_field_type_id = self.allowable_field_types[0].id
            field_value.allowable_field_type = self.allowable_field_types[0]
        if parent:
            field_value.set_parent(parent)
        return field_value


@add_schema
class FieldValue(FieldMixin, JSONSaveMixin, JSONDeleteMixin, ModelBase):
    """A FieldValue model. One of the more complex models.

    .. versionchanged:: 0.1.2     FieldValues no longer
    have'wires_as_source' or 'wires_as_dest' fields.    Wires may only
    be accessed via plans only or via the FieldValue instance method
    'get_wires,' which accesses the FieldValues operation and its Plan
    to obtain wires.
    """

    fields = dict(
        # FieldValue relationships
        field_type=HasOne("FieldType"),
        allowable_field_type=HasOne("AllowableFieldType"),
        array=fields.Field(),
        # array=Function(lambda fv: fv.array, callback_args=(fields.Callback.SELF,)),
        item=HasOne("Item", ref="child_item_id"),
        sample=HasOne("Sample", ref="child_sample_id"),
        object_type_id=Raw(),
        object_type=HasOne("ObjectType"),
        operation=HasOne("Operation", callback="find_field_parent", ref="parent_id"),
        parent_sample=HasOne("Sample", callback="find_field_parent", ref="parent_id"),
        sid=Function("get_sid"),
        child_sample_name=Function(
            lambda fv: fv.sid, callback_args=(fields.Callback.SELF,)
        ),
        wires_as_source=HasMany("Wire", ref="from_id"),
        wires_as_dest=HasMany("Wire", ref="to_id"),
        # allowable_child_types=Function('get_allowable_child_types'),
    )

    METATYPE = "field_type"

    def __init__(
        self,
        name=None,
        role=None,
        parent_class=None,
        parent_id=None,
        field_type=None,
        sample=None,
        value=None,
        item=None,
        container=None,
    ):
        """

        :param value:
        :type value:
        :param sample:
        :type sample:
        :param container:
        :type container:
        :param item:
        :type item:
        """
        super().__init__(
            name=name,
            role=role,
            parent_class=parent_class,
            parent_id=parent_id,
            field_type_id=None,
            field_type=field_type,
            child_sample_id=None,
            sample=sample,
            value=value,
            child_item_id=None,
            item=item,
            object_type=container,
            allowable_field_type_id=None,
            allowable_field_type=None,
            column=0,
            row=0,
        )

        if field_type is not None:
            self.set_field_type(field_type)

        if any([value, sample, item, container]):
            self._set_helper(
                value=value, sample=sample, item=item, object_type=container
            )

        if self.parent_class == "Operation" and not self.role:
            raise AquariumModelError(
                "FieldValue {} needs a role to be a field value for an operation".format(
                    self
                )
            )

    def get_sid(self):
        """The FieldValues sample identifier."""
        if self.sample is not None:
            return self.sample.identifier

    def get_wires(self):
        if not self.role or self.parent_class != "Operation":
            return None
        elif self.operation and self.operation.plan:
            wires = self.operation.plan.wires
            if self.role == "input":
                return [w for w in wires if w.does_wire_to(self)]
            elif self.role == "output":
                return [w for w in wires if w.does_wire_from(self)]
        return []

    @property
    def incoming_wires(self):
        if self.role == "input":
            self.get_wires()
        return []

    @property
    def outgoing_wires(self):
        if self.role == "output":
            self.get_wires()
        return []

    @property
    def successors(self):
        if self.outgoing_wires:
            return [x.destination for x in self.outgoing_wires]
        return []

    @property
    def predecessors(self):
        if self.incoming_wires:
            return [x.source for x in self.incoming_wires]
        return []

    def show(self, pre=""):
        if self.sample:
            if self.child_item_id:
                item = " item: {}".format(self.child_item_id) + " in {}".format(
                    self.item.object_type.name
                )
            else:
                item = ""
            print(
                "{}{}.{}:{}{}".format(pre, self.role, self.name, self.sample.name, item)
            )
        elif self.value:
            print("{}{}.{}:{}".format(pre, self.role, self.name, self.value))

    def reset(self):
        """Resets the inputs of the field_value."""
        self.value = None
        self.allowable_field_type_id = None
        self.allowable_field_type = None
        self.child_item_id = None
        self.item = None
        self.child_sample_id = None
        self.sample = None
        self.row = None
        self.column = None

    def _validate_types(self, name, x, expected_types):
        if x is not None and not any([issubclass(type(x), e) for e in expected_types]):
            raise AquariumModelError(
                "Cannot set FieldValue.{} with a {}. Expected types {}".format(
                    name, type(x), expected_types
                )
            )

    def _validate_value(self, value):
        if value is not None:
            choices = self.field_type.get_choices()
            if choices is not None:
                if value not in choices and str(value) not in choices:
                    raise AquariumModelError(
                        "Value '{}' not in list of field "
                        "type choices '{}'".format(value, choices)
                    )
        try:
            json.dumps(value)
        except TypeError as e:
            raise AquariumModelError(
                "Cannot set value '{}'. "
                "Value is not json serializable.".format(value)
            ) from e

    def _validate_item_and_container(self, item, container):
        self._validate_types("item", item, [Item, Collection])
        self._validate_types("container", container, [ObjectType])
        if item and container and item.object_type_id != container.id:
            raise AquariumModelError(
                "Item {} is not in container {}".format(item.id, str(container))
            )

    def _validate_sample(self, sample):
        self._validate_types("sample", sample, [Sample])
        if sample is not None and not issubclass(type(sample), Sample):
            raise AquariumModelError(
                "Cannot set sample with a type '{}'".format(type(sample))
            )

    def _validate_sample_and_item(self, sample, item):
        if sample and hasattr(sample, "id") and item:
            if not item.sample_id == sample.id:
                raise AquariumModelError(
                    "Cannot set FieldValue {}. Item {} is not a member of sample {}".format(
                        self, item, sample
                    )
                )

    def _validate_row_or_column(self, dim: int):
        if dim and not isinstance(dim, int):
            raise AquariumModelError("Row and columns must be integers not.")
        if dim is not None:
            if not self.field_type.part:
                raise AquariumModelError(
                    "Cannot set dimensions of a non-part for {} {}".format(
                        self.role, self.name
                    )
                )

    def _validate(self, value, sample, item, object_type, row, column):
        if row is not Null:
            self._validate_row_or_column(row)
        if column is not Null:
            self._validate_row_or_column(column)
        if value is not Null:
            self._validate_value(value)
        if sample is not Null:
            self._validate_sample(sample)

        if sample is Null and item is not Null:
            self._validate_sample_and_item(self.sample, item)
        if sample is not Null and item is not Null:
            self._validate_sample_and_item(sample, item)

        if item is Null and object_type is not Null:
            self._validate_item_and_container(self.item, object_type)
        if item is not Null and object_type is Null:
            self._validate_item_and_container(item, self.object_type)
        if item is not Null and object_type is not Null:
            self._validate_item_and_container(item, object_type)

    def _set_helper(
        self,
        value=Null,
        sample=Null,
        item=Null,
        object_type=Null,
        row=Null,
        column=Null,
    ):
        self._validate(value, sample, item, object_type, row, column)
        if row is not Null:
            self.row = row
        if column is not Null:
            self.column = column

        if value is not Null:
            self.value = value

        if item is not Null:
            self.item = item
            if hasattr(item, "id"):
                self.child_item_id = item.id
            if item and not sample:
                sample = item.sample
        if sample is not Null:
            self.sample = sample
            if hasattr(sample, "id"):
                self.child_sample_id = sample.id
        if object_type is not Null:
            self.object_type = object_type

    def valid_afts(self):
        afts = self.field_type.allowable_field_types
        if self.sample is not None:
            afts = filter_list(afts, sample_type_id=self.sample.sample_type_id)
        if self.object_type is not None:
            afts = filter_list(afts, object_type_id=self.object_type.id)
        if self.item is not None and self.item.object_type_id is not None:
            afts = filter_list(afts, object_type_id=self.item.object_type_id)
        return afts

    def _raise_aft_error(self):
        aft_list = []
        for aft in self.field_type.allowable_field_types:
            st = "none"
            ot = "none"
            if aft.object_type is not None:
                ot = aft.object_type.name
            if aft.sample_type is not None:
                st = aft.sample_type.name
            aft_list.append("{}:{}".format(st, ot))
        sid = "none"
        if self.sample is not None:
            sid = self.sample.sample_type.name
        oid = "none"
        if self.object_type is not None:
            oid = self.object_type.name
        msg = "No allowable field types found for {}:{}:{} using {}:{}."
        msg += " Available afts: {}"
        try:
            otname = self.operation.operation_type.name
        except AttributeError:
            otname = None
        raise AquariumModelError(
            msg.format(otname, self.role, self.name, sid, oid, ", ".join(aft_list))
        )

    def set_aft(self):
        afts = self.valid_afts()
        if len(afts) >= 1:
            self.set_allowable_field_type(afts[0])
        else:
            self._raise_aft_error()

    def set_value(
        self,
        value=Null,
        sample=Null,
        item=Null,
        object_type=Null,
        row=Null,
        column=Null,
        container=Null,
    ):
        if object_type and container:
            raise ValueError(
                "object_type and container (depreciated) cannot be "
                "both defined. Please define 'object_type'."
            )
        if container:
            warn(
                TridentDepreciationWarning(
                    "Use of 'container' is depreciated."
                    " Please use 'object_type' instead"
                )
            )
        if not object_type and container:
            object_type = container

        self._validate(value, sample, item, container, row, column)
        self._set_helper(
            value=value,
            sample=sample,
            item=item,
            object_type=object_type,
            row=row,
            column=column,
        )

        """Sets the value of a """
        if any([sample, object_type, item]):
            self.set_aft()
        return self

    def get_parent_key(self):
        if self.parent_class == "Operation":
            return "operation"
        elif self.parent_class == "Sample":
            return "parent_sample"
        else:
            raise AquariumModelError(
                "Parent class '{}' not recognized as a FieldValue parent.".format(
                    self.parent_class
                )
            )

    def get_parent(self):
        key = self.get_parent_key()
        return getattr(self, key)

    def set_parent(self, model):
        self.parent_id = model.id
        self.parent_class = model.__class__.__name__
        setattr(self, self.get_parent_key(), model)
        return self

    def set_operation(self, operation):
        self.parent_class = "Operation"
        if hasattr(operation, "id"):
            self.parent_id = operation.id
        self.operation = operation

    def set_sample(self, sample):
        self.parent_class = "Sample"
        if hasattr(sample, "id"):
            self.parent_id = sample.id
        self.sample = sample

    def set_field_type(self, field_type):
        """Sets properties from a field_type."""

        self.field_type = field_type
        self.field_type_id = field_type.id

        if field_type.name:
            self.name = field_type.name
        if field_type.parent_class:
            if field_type.parent_class == "OperationType":
                self.parent_class = "Operation"
            elif field_type.parent_class == "SampleType":
                self.parent_class = "Sample"
            else:
                raise AquariumModelError(
                    "FieldType parent_class '{}' not recognized".format(
                        field_type.parent_class
                    )
                )
        if field_type.role:
            self.role = field_type.role

    def set_allowable_field_type(self, allowable_field_type):
        self.allowable_field_type = allowable_field_type
        self.allowable_field_type_id = allowable_field_type.id

    def choose_item(self, first=True):
        """Set the item associated with the field value."""
        index = 0
        if not first:
            index = -1
        items = self.compatible_items()
        if items is not None and len(items) > 0:
            item = items[index]
            self.set_value(item=item)
            return item
        return None

    def compatible_items(self):
        """Find items compatible with the field value."""
        return self.session.utils.compatible_items(
            self.sample.id, self.allowable_field_type.object_type_id
        )

    def __str__(self):
        return self._to_str("id", "name", "role")



"""Planner."""
import itertools
import random
import webbrowser
from collections import defaultdict
from copy import deepcopy
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type
from typing import Union

import networkx as nx

from .plan_optimizer import PlanOptimizer
from pydent.aqsession import AqSession
from pydent.base import ModelBase
from pydent.exceptions import PlannerException
from pydent.exceptions import PlannerVerificationException
from pydent.exceptions import TridentBaseException
from pydent.models import AllowableFieldType
from pydent.models import FieldType
from pydent.models import FieldValue
from pydent.models import Item
from pydent.models import ObjectType
from pydent.models import Operation
from pydent.models import OperationType
from pydent.models import Plan
from pydent.models import Sample
from pydent.models import Wire
from pydent.planner.graph import PlannerGraph
from pydent.planner.graph import PlannerLayout
from pydent.planner.utils import _id_getter
from pydent.planner.utils import arr_to_pairs
from pydent.planner.utils import get_subgraphs
from pydent.utils import empty_copy
from pydent.utils import logger
from pydent.utils import make_async
from pydent.utils import QueryBuilder

QuickWireTargetType = Union[FieldValue, Operation, Tuple[Operation, str]]
UnresolvedWireTarget = Union[
    Operation, Tuple[Operation, str], str, FieldType, OperationType, FieldValue
]


def plan_verification_wrapper(fxn: Callable):
    """A wrapper that verifies that all FieldValues or Operations passed as
    arguments exist in the plan."""

    @wraps(fxn)
    def wrapper(self, *args, **kwargs):
        if not issubclass(self.__class__, Planner):
            raise ValueError(
                "Cannot apply 'verify_plan_models' to a non-planner instance."
            )
        for arg in args:
            if issubclass(arg.__class__, FieldValue):
                fv = arg
                if not self._contains_op(fv.operation):
                    fv_ref = "{} {}".format(fv.role, fv.name)
                    msg = 'FieldValue "{}" not found in planner.'
                    raise PlannerVerificationException(msg.format(fv_ref))
            elif issubclass(arg.__class__, Operation):
                op = arg
                if not self._contains_op(op):
                    op_ref = "{}".format(op.operation_type.name)
                    msg = 'Operation "{}" not found in planner.'
                    raise PlannerVerificationException(msg.format(op_ref))
        return fxn(self, *args, **kwargs)

    return wrapper


class AFTMatcher:
    @staticmethod
    def _resolve_to_field_types(
        model: Union[FieldValue, Operation, Tuple[Operation, str], FieldType],
        role: str = None,
    ):
        if isinstance(model, FieldValue):
            if model.role == role:
                return [model.field_type]
            else:
                msg = (
                    "Planner attempted to find matching"
                    " allowable_field_types for"
                    " an input FieldValue but found an output FieldValue"
                )
                raise PlannerException(msg)
        elif isinstance(model, Operation):
            return [ft for ft in model.get_field_types() if ft.role == role]
        elif isinstance(model, Tuple):
            op, ftname = model
            return [
                ft
                for ft in op.get_field_types()
                if ft.role == role and ft.name == ftname
            ]
        elif isinstance(model, FieldType):
            return [model]
        else:
            raise PlannerException(
                "Cannot resolve inputs, type must be a FieldValue, Operation, (Operation, category) or FieldType, not a "
                '"{}"'.format(type(model))
            )

    @classmethod
    def _collect_matching_afts(
        cls, source: QuickWireTargetType, destination: QuickWireTargetType
    ) -> Tuple[
        List[Tuple[AllowableFieldType, AllowableFieldType]],
        List[AllowableFieldType],
        List[AllowableFieldType],
    ]:
        """

        :param source: a field value or operation source
        :type source: FieldValue|Operation
        :param destination: a field value or operation destination
        :type destination: FieldValue|Operation
        :return: tuple of matching allowable_field_type (aft) pairs, matching
            field_value inputs and matching field_value outputs
        :rtype: tuple
        """
        """Find matching AllowableFieldTypes"""
        dest_fts = cls._resolve_to_field_types(destination, role="input")
        src_fts = cls._resolve_to_field_types(source, role="output")

        matching_afts = []
        matching_inputs = []
        matching_outputs = []
        for src in src_fts:
            for dest in dest_fts:
                io_matching_afts = cls._find_matching_afts(src, dest)
                if len(io_matching_afts) > 0:
                    if dest not in matching_inputs:
                        matching_inputs.append(dest)
                    if src not in matching_outputs:
                        matching_outputs.append(src)
                matching_afts += io_matching_afts
        return matching_afts, matching_inputs, matching_outputs

    @staticmethod
    def _find_matching_afts(
        src_ft: FieldType, dest_ft: FieldType
    ) -> List[Tuple[AllowableFieldType, AllowableFieldType]]:
        """Finds matching afts between two FieldTypes."""
        afts = []
        src_afts = src_ft.allowable_field_types
        dest_afts = dest_ft.allowable_field_types

        # check whether the field_type handles collections
        input_handles_collections = dest_ft.part is True
        output_handles_collections = dest_ft.part is True
        if input_handles_collections != output_handles_collections:
            return []

        for dest_aft in dest_afts:
            for src_aft in src_afts:
                out_object_type_id = src_aft.object_type_id
                in_object_type_id = dest_aft.object_type_id
                out_sample_type_id = src_aft.sample_type_id
                in_sample_type_id = dest_aft.sample_type_id
                if (
                    out_object_type_id == in_object_type_id
                    and out_sample_type_id == in_sample_type_id
                ):
                    afts.append((src_aft, dest_aft))
        return afts


class Planner(AFTMatcher):
    """A user-interface for making experimental plans and layouts."""

    class ITEM_SELECTION_PREFERENCE:
        """Item selection options."""

        ANY = "ANY"  #: pick the first item that matches the set field_value
        RESTRICT = "RESTRICT"  #: restrict to the currently set allowable_field_type
        PREFERRED = (
            "PREFERRED"
        )  #: (default) pick the item that matches the currently set aft, else
        #: pick ANY item that matches the set field_value
        RESTRICT_TO_ONE = (
            "RESTRICT TO ONE"
        )  #: will not select item if its being used in another active operation
        RESTRICT_TO_ONE_ON_SERVER = (
            "RESTRICT TO ONE ON SERVER"
        )  #: will restrict to a single item across the plans on the server.
        _DEFAULT = PREFERRED  #: default item selection preference
        _CHOICES = [ANY, RESTRICT, PREFERRED, RESTRICT_TO_ONE]  #: all choices

    class ITEM_ORDER_PREFERENCE:
        """Item order options."""

        FIRST = "FIRST"  #: select first item
        LAST = "LAST"  #: select last item
        RANDOM = "RANDOM"  #: select random item
        _DEFAULT = LAST  #: the default choice
        _CHOICES = [FIRST, LAST, RANDOM]  #: all choices

    def __init__(self, session_or_plan=None, plan_id=None):
        self._plan = None
        if issubclass(type(session_or_plan), AqSession):
            # initialize with session
            self.session = session_or_plan

            if plan_id is not None:
                # load an existing plan
                plan = self.browser.find(plan_id, "Plan")
                if plan is None:
                    raise PlannerException(
                        "Could not find plan with id={}".format(plan_id)
                    )
                self.plan = plan
            else:
                # create a new plan
                self.plan = self.session.Plan.new()
        elif issubclass(type(session_or_plan), Plan):
            # initialize with Plan
            plan = session_or_plan
            self.session = plan.session
            self.plan = plan
        self.logger = logger(self)

    @property
    def browser(self):
        return self.session.browser

    @classmethod
    def _check_plans_for_single_session(cls, models):
        session_ids = {id(m.session) for m in models}
        if len(session_ids) > 1:
            raise PlannerException("Plans have different session ids")
        if models:
            return models[0].session

    @staticmethod
    def _cache_query():
        return {
            "operations": {
                "field_values": {
                    "wires_as_dest": {
                        "source": "operation",
                        "destination": "operation",
                    },
                    "wires_as_source": {
                        "source": "operation",
                        "destination": "operation",
                    },
                    "sample": [],
                    "item": [],
                    "operation": [],
                    "field_type": [],
                },
                "plan_associations": ["plan"],
                "operation_type": {"field_types": []},
            }
        }

    @property
    def plan(self):
        return self._plan

    @plan.setter
    def plan(self, new_plan: Plan):
        self._plan = new_plan
        self.cache()

    @classmethod
    def cache_plans(cls, browser, plans: Iterable[Plan]) -> Iterable[Plan]:
        browser.get(plans, cls._cache_query())
        for plan in plans:
            wire_dict = {}
            if plan.operations:
                for op in plan.operations:
                    for fv in op.field_values:
                        for w in fv.wires_as_dest:
                            wire_dict[w._primary_key] = wire_dict.get(w._primary_key, w)
                        for w in fv.wires_as_source:
                            wire_dict[w._primary_key] = wire_dict.get(w._primary_key, w)
            plan.wires = list(wire_dict.values())
        return plans

    def cache(self):
        self.cache_plans(self.browser, [self.plan])

    @property
    def name(self):
        return self.plan.name

    @name.setter
    def name(self, value):
        self.plan.name = value

    @property
    def url(self):
        url = self.session.url + r"/plans?plan_id={}".format(self.plan.id)
        url = url.replace(r"//", r"/")
        return url

    def open(self):
        webbrowser.open(self.url)

    def create(self):
        """Create the plan on Aquarium."""
        if self.plan.id:
            raise PlannerException(
                "Cannot create plan since it already exists on the server (plan_id={plan_id})"
                " Did you mean .{save}() push an update to the server plan? You"
                " can also create a copy of the new plan by calling .{replan}().".format(
                    plan_id=self.plan.id,
                    save=self.save.__name__,
                    replan=self.replan.__name__,
                )
            )

        self.plan.create()

    # TODO: fix this 'set_timeout' to not be global
    def save(self):
        """Saves the Plan to the Aquarium server.

        :return: None
        """
        if not self.plan.id:
            self.create()
        else:
            self.update()

    def update(self):
        """Update the plan on Aquarium."""
        self.plan.save()
        return self.plan

    def delete(self):
        """Delete the plan on the Aquarium server. Make an annonymized copy of
        the plan stored in Trident.

        .. versionadded:: 0.1.5a10     Added delete method
        """
        self.plan.delete()
        self._plan = self.plan.copy()

    def create_operation_by_type(
        self, ot: OperationType, status: str = "planning"
    ) -> Operation:
        op = ot.instance()
        op.status = status
        self.plan.add_operation(op)
        self.logger.debug("{} created".format(ot.name))
        return op

    def create_operation_by_type_id(self, ot_id: Union[int, str]) -> Operation:
        # ot = self.browser.find('OperationType', ot_id)
        ot = self.session.OperationType.find(ot_id)
        return self.create_operation_by_type(ot)

    def create_operation_by_name(
        self, operation_type_name: str, category: str = None
    ) -> Operation:
        """Adds a new operation to the plan."""
        query = {"deployed": True, "name": operation_type_name}
        if category is not None:
            query["category"] = category
        ots = self.session.OperationType.where(query)
        if len(ots) > 1:
            msg = (
                'Found more than one OperationType for query "{}". Have you tried '
                "specifying the category?"
            )
            raise PlannerException(msg.format(query))
        if ots is None or len(ots) == 0:
            msg = 'Could not find deployed OperationType "{}".'
            raise PlannerException(msg.format(operation_type_name))
        return self.create_operation_by_type(ots[0])

    def new_op(self, name_id_or_ot: Union[str, int, OperationType]) -> Operation:
        if isinstance(name_id_or_ot, str):
            return self.create_operation_by_name(name_id_or_ot)
        elif isinstance(name_id_or_ot, int):
            return self.create_operation_by_type_id(name_id_or_ot)
        elif issubclass(type(name_id_or_ot), OperationType):
            return self.create_operation_by_type(name_id_or_ot)
        else:
            raise PlannerException(
                "Expected a string, integer, or OperationType, "
                "not a {}".format(type(name_id_or_ot))
            )

    @staticmethod
    def _model_are_equal(model1: ModelBase, model2: ModelBase) -> bool:
        mid1 = model1.id or model1._primary_key
        mid2 = model2.id or model2._primary_key
        return mid1 == mid2
        # if model1.id is None and model2.id is None:
        #     if model1._primary_key == model2._primary_key:
        #         return True
        # elif model1.id == model2.id:
        #     return True
        # return False

    def get_op(self, id: Union[str, int]) -> Union[Operation, None]:
        for op in self.plan.operations:
            if op.id == id:
                return op
            elif op._primary_key == id:
                return op

    @plan_verification_wrapper
    def get_wire(self, fv1: FieldValue, fv2: FieldValue) -> Union[None, Wire]:
        for wire in self.plan.wires:
            if self._model_are_equal(wire.source, fv1) and self._model_are_equal(
                wire.destination, fv2
            ):
                self.logger.debug("found wire from {} to {}".format(fv1.name, fv2.name))
                return wire

    @plan_verification_wrapper
    def remove_wire(self, fv1: FieldValue, fv2: FieldValue) -> Wire:
        """Removes a wire between two field values from a plan.

        :param fv1:
        :param fv2:
        :return:
        """
        wire = self.get_wire(fv1, fv2)
        wires = list(self.plan.wires)
        if wire:
            self.logger.debug("removing wire from {} to {}".format(fv1.name, fv2.name))
            # wires_as_source = fv1.wires_as_source
            # wires_as_dest = fv2.wires_as_dest
            #
            # wires_as_source.remove(wire)
            # wires_as_dest.remove(wire)

            # fv1.wires_as_source = wires_as_source
            # fv2.wires_as_dest = wires_as_dest

            wires.remove(wire)
            self.plan.wires = wires
        return wire

    @plan_verification_wrapper
    def get_outgoing_wires(self, fv: FieldValue) -> List[Wire]:
        wires = []
        for wire in self.plan.wires:
            if self._model_are_equal(wire.source, fv):
                wires.append(wire)
        return wires

    @plan_verification_wrapper
    def get_incoming_wires(self, fv: FieldValue) -> List[Wire]:
        wires = []
        for wire in self.plan.wires:
            if self._model_are_equal(wire.destination, fv):
                wires.append(wire)
        return wires

    def get_fv_successors(self, fv: FieldValue) -> List[FieldValue]:
        fvs = []
        for wire in self.get_outgoing_wires(fv):
            fvs.append(wire.destination)
        return fvs

    def get_fv_predecessors(self, fv: FieldValue) -> List[FieldValue]:
        fvs = []
        for wire in self.get_incoming_wires(fv):
            fvs.append(wire.source)
        return fvs

    def get_op_successors(self, op: Operation) -> List[Operation]:
        ops = []
        for output in op.outputs:
            ops += [fv.operation for fv in self.get_fv_successors(output)]
        return ops

    def get_op_predecessors(self, op: Operation) -> List[Operation]:
        ops = []
        for input in op.inputs:
            ops += [fv.operation for fv in self.get_fv_predecessors(input)]
        return ops

    def quick_create_operation_by_name(self, otname: str) -> Operation:
        try:
            return self.create_operation_by_name(*otname)
        except TypeError:
            return self.create_operation_by_name(otname)

    def quick_create_and_wire(self, otname1, otname2, fvnames=None):
        self.quick_create_operation_by_name(otname1)
        self.quick_create_operation_by_name(otname2)
        return self.quick_wire_by_name(otname1, otname2)

    def _contains_op(self, op):
        if op in self.plan.operations:
            return True
        else:
            plan_operation_ids = [x.id for x in self.plan.operations]
            return op.id is not None and op.id in plan_operation_ids

    @plan_verification_wrapper
    def _resolve_to_wire_target(
        self,
        unresolved: UnresolvedWireTarget,
        category: str = None,
        return_type: Union[Type[Operation], None] = None,
    ) -> Union[Operation, FieldValue]:
        if isinstance(unresolved, tuple):
            resolved = self.create_operation_by_name(
                unresolved[0], category=unresolved[1]
            )
        elif isinstance(unresolved, str):
            resolved = self.create_operation_by_name(unresolved, category=category)
        elif isinstance(unresolved, OperationType):
            resolved = self.create_operation_by_type(unresolved, category=category)
        elif isinstance(unresolved, FieldType):
            op = self.create_operation_by_type_id(unresolved.parent_id)
            ft = unresolved
            if unresolved.role == "input":
                if unresolved.array:
                    resolved = self._select_empty_input_array(op, ft.name)
                else:
                    resolved = op.input(ft.name)
            elif unresolved.role == "output":
                if unresolved.array:
                    resolved = self._select_empty_input_array(op, ft.name)
                else:
                    resolved = op.output(ft.name)
        elif isinstance(unresolved, FieldValue) or isinstance(unresolved, Operation):
            resolved = unresolved
        else:
            raise TypeError(unresolved)

        if return_type is None:
            return resolved
        elif Operation:
            if isinstance(resolved, Operation):
                return resolved
            else:
                return resolved.operation

    @property
    def operations(self) -> List[Operation]:
        """Return list of :class:`operations <pydent.models.Operation>` in the
        plan.

        .. versionadded: 0.1.5a10
            Added operations property to planner

        :return: list of :class:`operations <pydent.models.Operation>`
        """
        return self.plan.operations

    @operations.setter
    def operations(self, ops: List[Operation]):
        self.plan.operations = ops

    def chain(
        self,
        *op_or_otnames: Tuple[UnresolvedWireTarget, ...],
        category: str = None,
        return_as_dict: bool = False,
    ) -> List[Operation]:
        """Creates a chain of operations by *guessing* wires between operations
        based on the AllowableFieldTypes between the inputs and outputs of each
        operation type. Sample inputs and outputs will be set along the wire if
        possible.

        e.g.

        .. code::

            # create four new operations based on their OperationType names
            planner.chain("Make PCR Fragment", "Run Gel",
                                      "Extract Gel Slice", "Purify Gel Slice")

            # create four new operations based on their OperationType names by
            # finding OperationTypes only in the "Cloning" category
            planner.chain("Make PCR Fragment", "Run Gel",
                                      "Extract Gel Slice", "Purify Gel Slice",
                                      category="Cloning")

            # create four new operations based on their OperationType names by
            # finding OperationTypes only in the "Cloning" category,
            # except find "Make PCR Fragment" in the "Cloning Sandbox" category
            planner.chain(("Make PCR Fragment", "Cloning Sandbox"),
                                       "Run Gel", "Extract Gel Slice",
                                       "Purify Gel Slice", category="Cloning")

            # create and wire new operations to an existing operations while
            # routing samples
            pcr_op = planner.create_operation_by_name("Make PCR Fragment")
            planner.set_field_value(pcr_op.outputs[0], sample=my_sample)
            new_ops = planner.chain(pcr_op, "Run Gel")
            run_gel = new_ops[1]
            planner.chain("Pour Gel", run_gel)
        """
        self.logger.debug("QUICK CREATE CHAIN {}".format(op_or_otnames))
        targets = [
            self._resolve_to_wire_target(n, category=category) for n in op_or_otnames
        ]
        if any([op for op in targets if op is None]):
            raise PlannerException("Could not find some operations: {}".format(targets))
        pairs = arr_to_pairs(targets)
        for src, dest in pairs:
            self.quick_wire(src, dest)
        if return_as_dict:
            return_dict = {}
            for op_or_fv in targets:
                return_dict.setdefault(op_or_fv.operation_type.name, []).append(
                    op_or_fv
                )
            return return_dict
        return targets

    def _select_empty_input_array(
        self, op: Operation, fvname: str
    ) -> Union[None, FieldValue]:
        """Selects the first 'empty' (i.e. field_values with no Sample set)
        field value in the :class:`FieldValue` array. Returns None if the
        FieldType is not an array. If there are no current 'empty' field
        values, a new one is instantiated and returned.

        :param op: Operation
        :param fvname: FieldType/FieldValue name of the array
        :return:
        """
        field_type = op.operation_type.field_type(fvname, "input")
        if field_type.array:
            existing_fvs = op.input_array(fvname)
            for fv in existing_fvs:
                if fv.sample is None and not self.get_incoming_wires(fv):
                    return fv
            fv = op.new_field_value(field_type.name, "input")
            return fv

    # TODO: way to select preference for afts in quick_wire?
    @plan_verification_wrapper
    def quick_wire(
        self,
        source: Union[Operation, FieldValue, Tuple[Operation, str]],
        destination: Union[Operation, FieldValue, Tuple[Operation, str]],
        strict: bool = False,
    ) -> Wire:
        """

        :param source: field_value or operation source
        :type source: FieldValue|Operation
        :param destination: field_value or operation destination
        :type destination: FieldValue|Operation
        :param strict: will raise error if there is is ambiguous wiring between the
            source and destination
        :type strict: bool
        :return: created wire
        :rtype: Wire
        """
        if isinstance(source, tuple):
            source, src_fvname = source
            src_target = source.operation_type.field_type(src_fvname, role="output")
        else:
            src_target = source
        if isinstance(destination, tuple):
            destination, dest_fvname = destination
            dest_target = destination.operation_type.field_type(
                dest_fvname, role="input"
            )
        else:
            dest_target = destination
        afts, model_inputs, model_outputs = self._collect_matching_afts(
            src_target, dest_target
        )

        # TODO: only if matching FVs are ambiguous raise Exception
        # TODO: automatically wire to empty FV if they are equivalent
        if (len(model_inputs) > 1 or len(model_outputs) > 1) and strict:
            raise PlannerException(
                "Cannot quick wire. Ambiguous wiring between inputs [{}] for {} and "
                "outputs [{}] for {}\nInstead, try to use `add_wire` method to wire "
                "together two FieldValues.".format(
                    ", ".join([ft.name for ft in model_inputs]),
                    model_inputs[0].operation_type.name,
                    ", ".join([ft.name for ft in model_outputs]),
                    model_outputs[0].operation_type.name,
                )
            )
        elif len(afts) > 0:
            for aft1, aft2 in afts:

                input_ft = aft2.field_type
                output_ft = aft1.field_type

                # input_fv = None

                if isinstance(source, FieldValue):
                    output_fv = source
                else:
                    output_fv = source.output(output_ft.name)

                if isinstance(destination, FieldValue):
                    input_fv = destination
                else:
                    if input_ft.array:
                        input_fv = self._select_empty_input_array(
                            destination, input_ft.name
                        )
                    else:
                        input_fv = destination.input(input_ft.name)

                return self.add_wire(output_fv, input_fv)

        elif len(afts) == 0:
            raise PlannerException(
                "Cannot quick wire. No possible wiring found between inputs {} and {}".format(
                    source, destination
                )
            )

    def quick_wire_by_name(self, otname1: str, otname2: str) -> Wire:
        """Wires together the last added operations."""
        op1 = self.get_op_by_name(otname1)[-1]
        op2 = self.get_op_by_name(otname2)[-1]
        return self.quick_wire(op1, op2)

    def clean_wires(self):
        wires_by_id = {}
        for wire in self.plan.wires:
            if wire.source is not None and wire.destination is not None:
                _id = "{}_{}".format(
                    wire.source._primary_key, wire.destination._primary_key
                )
                wires_by_id[_id] = wire
        for op in self.plan.operations:
            for fv in op.field_values:
                wires_as_source = fv.wires_as_source
                wires_as_dest = fv.wires_as_dest

                if wires_as_source is None:
                    wires_as_source = []
                if wires_as_dest is None:
                    wires_as_dest = []

                fv.wires_as_source = list(set(wires_as_source))
                fv.wires_as_dest = list(set(wires_as_dest))
        self.plan.wires = list(wires_by_id.values())

    def remove_operations(self, ops: Iterable[Operation]):
        self.clean_wires()
        operations = self.plan.operations
        wires = set(self.plan.wires)
        wires_to_remove = set()

        for op in ops:
            operations.remove(op)
            for fv in op.field_values:
                wires_to_remove = wires_to_remove.union(set(fv.wires_as_source))
                wires_to_remove = wires_to_remove.union(set(fv.wires_as_dest))

        wires = list(wires.difference(wires_to_remove))

        self.plan.operations = operations
        self.plan.wires = wires

    # TODO: resolve afts if already set...
    # TODO: clean up _set_wire
    def _set_wire(
        self,
        src_fv: FieldValue,
        dest_fv: FieldValue,
        preference: str = "source",
        setter: Union[None, Callable] = None,
    ):

        if len(self.get_incoming_wires(dest_fv)) > 0:
            raise PlannerException(
                'Cannot wire to destination FieldValue "{}" already has an incoming'
                " wire and inputs"
                " can only have one incoming wire. Please remove wire"
                " using 'planner.remove_wire(src_fv, dest_fv)' before setting.".format(
                    dest_fv.name
                )
            )

        # first collect any matching allowable field types between the field values
        aft_pairs = self._collect_matching_afts(src_fv, dest_fv)[0]

        if len(aft_pairs) == 0:
            raise PlannerException(
                'Cannot wire "{}" to "{}". No allowable field types match.'.format(
                    src_fv.name, dest_fv.name
                )
            )

        # select the first aft
        default_aft_ids = [
            src_fv.allowable_field_type_id,
            dest_fv.allowable_field_type_id,
        ]
        pref_index = 0
        if preference == "destination":
            pref_index = 1
        selected_aft_pair = aft_pairs[0]
        for pair in aft_pairs:
            if pair[pref_index].id == default_aft_ids[pref_index]:
                selected_aft_pair = pair
        assert (
            selected_aft_pair[0].object_type_id == selected_aft_pair[1].object_type_id
        )

        # resolve sample
        samples = [src_fv.sample, dest_fv.sample]
        samples = [s for s in samples if s is not None]

        if setter is None:
            setter = self.set_field_value

        if len(samples) > 0:
            selected_sample = samples[0]

            # filter afts by sample_type_id
            aft_pairs = [
                aft
                for aft in aft_pairs
                if aft[0].sample_type_id == selected_sample.sample_type_id
            ]
            selected_aft_pair = aft_pairs[0]

            if len(aft_pairs) == 0:
                raise PlannerException(
                    "No allowable_field_types were found for FieldValues {} & {} for"
                    " Sample {}".format(src_fv.name, dest_fv.name, selected_sample.name)
                )

            setter(
                src_fv,
                sample=selected_sample,
                object_type=selected_aft_pair[0].object_type,
            )
            setter(
                dest_fv,
                sample=selected_sample,
                object_type=selected_aft_pair[0].object_type,
            )

            assert (
                selected_aft_pair[0].sample_type_id
                == selected_aft_pair[1].sample_type_id
            )
            assert selected_aft_pair[0].sample_type_id == src_fv.sample.sample_type_id
            assert selected_aft_pair[0].sample_type_id == dest_fv.sample.sample_type_id
        else:
            setter(src_fv, object_type=selected_aft_pair[0].object_type)
            setter(dest_fv, object_type=selected_aft_pair[0].object_type)

            # # set the sample (and allowable_field_type)
            # if src_fv.sample is not None and (
            #     dest_fv.sample is None or dest_fv.sample.id != src_fv.sample.id
            # ):
            #     setter(
            #         dest_fv,
            #         sample=src_fv.sample,
            #         object_type=selected_aft_pair[0].object_type,
            #     )
            # elif dest_fv.sample is not None and (
            #     src_fv.sample is None or dest_fv.sample.id != src_fv.sample.id
            # ):
            #     setter(
            #         src_fv,
            #         sample=dest_fv.sample,
            #         object_type=selected_aft_pair[0].object_type,
            #     )

    @plan_verification_wrapper
    def add_wire(self, fv1: FieldValue, fv2: FieldValue) -> Wire:
        """Add wire from one FieldValue to another FieldValue.

        :param fv1: source FieldValue
        :param fv2: destination FieldValue
        :return: the new Wire
        """
        wire = self.get_wire(fv1, fv2)
        if wire is None:
            # wire does not exist, so create it
            self._set_wire(fv1, fv2)
            wire = self.plan.wire(fv1, fv2)
            self.logger.debug("wired {} to {}".format(fv1.name, fv2.name))
        return wire

    @classmethod
    def _routing_id(cls, fv: FieldValue) -> str:
        routing_id = "{}_{}".format(_id_getter(fv.operation), fv.field_type.routing)
        if fv.field_type.array and fv.role == "input":
            other_fvs = fv.operation.input_array(fv.name)
            for pos, other_fv in enumerate(other_fvs):
                if cls._model_are_equal(other_fv, fv):
                    routing_id += str(pos)
                    return routing_id
        return routing_id

    @staticmethod
    def get_sample_routing_of_operation(op: Operation) -> Dict[str, List[FieldValue]]:
        routing_dict = {}
        for fv in op.field_values:
            routing = fv.field_type.routing
            routing_fvs = routing_dict.get(routing, [])
            routing_fvs.append(fv)
            routing_dict[fv.field_type.routing] = routing_fvs
        return routing_dict

    def _routing_graph(self):
        """Get sample routing graph.

        A property of a valid plan is that any two routing nodes that
        are connected by an edge must have the same sample. Edges are
        treated as 'sample wires'
        """
        G = nx.DiGraph()
        for op in self.plan.operations:
            for fv in op.field_values:
                G.add_node(self._routing_id(fv), fv=fv)

        for w in self.plan.wires:
            src_id = self._routing_id(w.source)
            dest_id = self._routing_id(w.destination)

            G.add_node(src_id, fv=w.source)
            G.add_node(dest_id, fv=w.destination)

            G.add_edge(src_id, dest_id, wire=w)
        return G

    # TODO: Support for row and column
    # TODO: routing dict does not work with input arrays (it groups them ALL together)
    @plan_verification_wrapper
    def set_field_value(
        self,
        field_value: FieldValue,
        sample: Union[None, Sample] = None,
        item: Union[None, Item] = None,
        object_type: Union[None, ObjectType] = None,
        value: Any = None,
        row: Union[None, int] = None,
        column: Union[None, int] = None,
    ):
        """.. versionchanged:: 0.1.5a15 Argument `container` changed to \
        `object_type` to keep consistency with other code.

        :param field_value:
        :param sample:
        :param item:
        :param object_type:
        :param value:
        :param row:
        :param column:
        :return:
        """
        self.logger.debug(
            "setting field_value {} to {} - {} - {} - {}".format(
                field_value.name, sample, item, object_type, value
            )
        )
        field_value.set_value(
            sample=sample,
            item=item,
            container=object_type,
            value=value,
            row=None,
            column=None,
        )
        if not field_value.field_type.array:
            routing = field_value.field_type.routing
            fvs = self.get_sample_routing_of_operation(field_value.operation)[routing]
            if field_value.field_type.ftype == "sample":
                for fv in fvs:
                    fv.set_value(sample=sample, row=row, column=column)
        return field_value

    def set_input_field_value_array(
        self,
        op: Operation,
        field_value_name: str,
        sample: Union[None, Sample] = None,
        item: Union[None, Item] = None,
        container: Union[None, ObjectType] = None,
        value: Any = None,
        row: Union[None, int] = None,
        column: Union[None, int] = None,
    ):
        """Finds the first 'empty' (no incoming wires and no sample set) field
        value and set the field value. If there are no empty field values in
        the array, create a new field value and set that one.

        :param op: operation
        :type op: Operation
        :param field_value_name: field value name
        :type field_value_name: basestring
        :param sample: the sample to set
        :type sample: Sample
        :param item: the item to set
        :type item: Item
        :param container: the container type to set
        :type container: ObjectType
        :return: the set field value
        :rtype: FieldValue
        """
        input_fv = self._select_empty_input_array(op, field_value_name)
        return self.set_field_value(
            input_fv,
            sample=sample,
            item=item,
            object_type=container,
            row=row,
            column=column,
            value=value,
        )

    def set_field_value_and_propogate(
        self,
        field_value: FieldValue,
        sample: Union[None, Sample] = None,
        item: Union[None, Item] = None,
        container: Union[None, ObjectType] = None,
        value: Any = None,
        row: Union[None, int] = None,
        column: Union[None, int] = None,
    ):
        if item:
            if sample:
                sid1 = sample.id or sample._primary_key
                sid2 = item.sample_id or item.sample.id or item.sample._primary_key
                if not sid1 == sid2:
                    raise PlannerException(
                        "Provided sample '{}' and item '{}' cannot"
                        " be different".format(sample, item)
                    )
            else:
                sample = item.sample
            self.set_field_value(field_value, item=item, row=row, column=column)
        routing_graph = self._routing_graph()
        routing_id = self._routing_id(field_value)
        subgraph = nx.bfs_tree(routing_graph.to_undirected(as_view=True), routing_id)
        for node in subgraph:
            n = routing_graph.nodes[node]
            fv = n["fv"]
            self.set_field_value(fv, sample=sample, object_type=container, value=value)
        return field_value

    @staticmethod
    @make_async
    def _filter_by_lambdas(
        models: List[Any], lambda_arr: Iterable[Callable]
    ) -> List[Any]:
        """Filters models by a array of lambdas."""
        arr = models[:]
        _arr = []
        while arr:
            m = arr.pop()
            for fxn in lambda_arr:
                try:
                    if fxn(m):
                        _arr.append(m)
                except TypeError or ValueError:
                    continue
        return _arr

    def _item_preference_query(
        self, sample: Sample, field_value: FieldValue, item_preference: str
    ) -> dict:
        if item_preference not in self.ITEM_SELECTION_PREFERENCE._CHOICES:
            raise PlannerException(
                'Item selection preference "{}" not recognized'
                " Please select from one of {}".format(
                    item_preference, ",".join(self.ITEM_SELECTION_PREFERENCE._CHOICES)
                )
            )
        query = {"sample_id": sample.id}

        # restrict allowable_field_types based on preference
        afts = list(field_value.field_type.allowable_field_types)

        if item_preference in [
            self.ITEM_SELECTION_PREFERENCE.RESTRICT,
            self.ITEM_SELECTION_PREFERENCE.RESTRICT_TO_ONE,
        ]:
            afts = [field_value.allowable_field_type]
        if item_preference == self.ITEM_SELECTION_PREFERENCE.PREFERRED:
            afts = sorted(
                afts,
                reverse=True,
                key=lambda aft: aft.sample_type_id == sample.sample_type_id,
            )

        query.update({"object_type_id": [aft.object_type_id for aft in afts]})
        return query

    def reserved_items(
        self, items: List[Item], search_server: bool = False
    ) -> Dict[int, List[FieldValue]]:
        """Returns a dictionary of item_ids and the array of field_values that
        use them."""
        browser = self.browser
        item_ids = [i.id for i in items]
        if search_server:
            server_fvs = browser.where(
                {"child_item_id": item_ids}, model_class="FieldValue"
            )
            browser.retrieve(server_fvs, "operation")
            server_fvs = [
                fv
                for fv in server_fvs
                if fv.operation and fv.operation and fv.operation.status != "planning"
            ]
        else:
            server_fvs = []

        these_fvs = []
        for op in self.plan.operations:
            for fv in op.inputs:
                these_fvs.append(fv)

        fvs = list(set(server_fvs + these_fvs))

        item_id_to_fv = defaultdict(list)
        for fv in fvs:
            if fv.child_item_id:
                item_id_to_fv[fv.child_item_id].append(fv)
        return item_id_to_fv

    @plan_verification_wrapper
    def get_available_items(
        self, field_value: FieldValue, item_preference: str
    ) -> List[Item]:
        sample = field_value.sample
        if not sample:
            return []

        query = self._item_preference_query(sample, field_value, item_preference)
        available_items = self.session.Item.where(query)
        available_items = [i for i in available_items if i.location != "deleted"]

        x = len(available_items)
        if item_preference == self.ITEM_SELECTION_PREFERENCE.RESTRICT_TO_ONE_ON_SERVER:
            reserved = self.reserved_items(available_items, search_server=True)
            available_items = [i for i in available_items if len(reserved[i.id]) == 0]
            self.logger.debug("{} items are reserved".format(x - len(available_items)))
        elif item_preference == self.ITEM_SELECTION_PREFERENCE.RESTRICT_TO_ONE:
            reserved = self.reserved_items(available_items, search_server=False)
            available_items = [i for i in available_items if len(reserved[i.id]) == 0]
            self.logger.debug("{} items are reserved".format(x - len(available_items)))
        return available_items

    def distribute_items_of_object_type(self, object_type: ObjectType):
        """Distribute items of a particular object_type across non-planning
        operations across existing plans and these operations in this Planner
        instance.

        E.g. distributing one-shot yeast competent cell aliquots
        """
        # collect items of that type used
        items = []
        browser = self.browser
        for op in self.plan.operations:
            for i in op.inputs:
                item = i.item
                if item:
                    if item.object_type_id == object_type.id:
                        items.append(item)

        # find all field_values that use this item
        item_id_to_fv = defaultdict(list)
        item_ids = [i.id for i in items]
        fvs = browser.where({"child_item_id": item_ids}, model_class="FieldValue")
        browser.retrieve(fvs, "operation")

        for fv in fvs:
            item_id_to_fv[fv.child_item_id].append(fv)

        distributed = []

        for item in items:
            fvs = item_id_to_fv[item.id]

            # filter by operations that are submitted
            fvs = [
                fv for fv in fvs if fv.operation and fv.operation.status != "planning"
            ]
            if len(fvs) > 1:
                available_items = browser.where(
                    {
                        "sample_id": item.sample_id,
                        "object_type_id": item.object_type_id,
                    },
                    model_class="Item",
                )
                available_items = [
                    item for item in available_items if item.location != "deleted"
                ]
                print(
                    "Item {} {} is used in {} inputs. There are {} available items.".format(
                        item.id, item.sample.name, len(fvs), len(available_items)
                    )
                )

                for fv, item in itertools.zip_longest(
                    fvs, available_items, fillvalue="&&&"
                ):
                    if fv != "&&&" and item != "&&&":
                        distributed.append(fv, item)

        for fv, item in distributed:
            fv.set_value(item=item)

    @plan_verification_wrapper
    def set_to_available_item(
        self,
        fv: FieldValue,
        order_preference: str = ITEM_ORDER_PREFERENCE._DEFAULT,
        filter_func: Union[None, Callable] = None,
        item_preference: str = ITEM_SELECTION_PREFERENCE._DEFAULT,
    ) -> Item:
        """Sets the item of the field value to the next available item. Setting
        recent=False will select the oldest item.

        :param fv: The field value to set.
        :type fv: FieldValue
        :param recent: whether to select the most recent item (recent=True is default)
        :type recent: bool
        :param filter_func: array of lambda to filter items by. E.g.
            'lambda_arr=[lambda x: x.get("concentration") > 10]'
        :type filter_func: list
        :param order_preference: The item order preference. Select from "LAST"
            (default), "FIRST", or "RANDOM". Selections
            also available in :class:`Planner.ITEM_ORDER_PREFERENCE`.
            Random will choose a random item from the available items found after
            all preferences and filters
        :type order_preference: basestring
        :param item_preference: The item selection preference. Select from items in
            :class:`Planner.ITEM_SELECTION_PREFERENCE`
            If "PREFERRED" (default), will select item that matches the
            allowable_field_value that is already set for the field_value, else select
            any item that matches the :class:`Sample` set :class:`FieldValue`. If
            "RESTRICT", will restrict selected item only to those that matches the
            current allowable_field_value that is set for the :class:`FieldValue`. If
            "ANY", will select any items that matches the :class:`Sample`.set for
            the field_value
        :type item_preference: basestring
        :return: The selected item (or None if no item set)
        :rtype: Item or None
        """
        sample = fv.sample
        item = None

        if sample is not None:
            available_items = self.get_available_items(fv, item_preference)
            if filter_func is not None:
                if not isinstance(filter_func, list):
                    filter_func = [filter_func]
                available_items = self._filter_by_lambdas(available_items, filter_func)

            if not available_items:
                return None

            available_items = sorted(available_items, key=lambda x: x.created_at)

            selection_index = -1
            if order_preference == self.ITEM_ORDER_PREFERENCE.FIRST:
                selection_index = 0
            elif order_preference == self.ITEM_ORDER_PREFERENCE.LAST:
                selection_index = -1
            elif order_preference == self.ITEM_ORDER_PREFERENCE.RANDOM:
                selection_index = random.randint(0, len(available_items) - 1)
            item = available_items[selection_index]
            fv.set_value(item=item)
        return item

    @plan_verification_wrapper
    def set_inputs_using_sample_properties(
        self,
        operation: Operation,
        sample: Sample,
        routing: dict = None,
        setter: Union[None, Callable] = None,
    ):
        """Map the sample field values to the operation inputs. Optionally, a
        routing dictionary may be passed to indicate the mapping between the
        sample field values and operation inputs.

        For example, the following would map the "Integrant" sample field value to the
        "Template" operation input, and so on...:

        .. code:: python

            routing={
                "Integrant": "Template",
                "QC_Primer1": "Forward Primer",
                "QC_Primer2": "Reverse Primer"
            }
        """
        if routing is None:
            routing = {ft.name: ft.name for ft in sample.sample_type.field_types}

        if setter is None:
            setter = self.set_field_value
        for sfv, ofv in routing.items():
            operation_input = operation.input(ofv)
            sample_property = sample.properties.get(sfv, None)
            # set operation input to the property sample
            if operation_input is not None and sample_property is not None:
                setter(operation_input, sample=sample_property)

    @plan_verification_wrapper
    def set_output_sample(
        self,
        fv: FieldValue,
        sample: Sample = None,
        routing: dict = None,
        setter: Union[None, Callable] = None,
    ):
        """Sets the output of the field value to the sample. If the field_value
        names between the Sample and the field_value's operation inputs, these
        will be set as well. Optionally, a routing dictionary may be passed to
        indicate the mapping between the sample field values and the operation
        field values.

        :param fv: the output field value
        :type fv:
        :param sample:
        :type sample:
        :param routing:
        :type routing:
        :return:
        :rtype:
        """
        if fv.role != "output":
            raise PlannerException(
                "Cannot output. FieldValue {} is a/an {}.".format(fv.name, fv.role)
            )
        if setter is None:
            setter = self.set_field_value
        setter(fv, sample=sample)
        op = fv.operation
        self.set_inputs_using_sample_properties(
            op, sample, routing=routing, setter=setter
        )

    @staticmethod
    def _json_update(model: ModelBase, **params: dict) -> ModelBase:
        """Temporary method to update."""
        aqhttp = model.session._http
        data = {"model": {"model": model.__class__.__name__}}
        data.update(model.dump(**params))
        model_data = aqhttp.create("json/save", json_data=data)
        model.reload(model_data)
        return model

    @classmethod
    def move_op(cls, op: Operation, plan_id: Union[str, id]):
        pa = op.plan_associations[0]
        pa.plan_id = plan_id
        op.parent_id = 0
        cls._json_update(op)
        cls._json_update(pa)

    def get_op_by_name(self, operation_type_name: str) -> Operation:
        """Find operations by their operation_type_name.

        :param operation_type_name: The operation type name
        :type operation_type_name: basestring
        :return: list of operations
        :rtype: list
        """
        return [
            op
            for op in self.plan.operations
            if op.operation_type.name == operation_type_name
        ]

    def replan(self) -> "Planner":
        """Replan the plan, by 'copying' the plan using the Aquarium server."""
        planner = self.__class__(self.session)
        planner.plan = self.plan.replan()
        return planner

    def annotate(
        self, markdown: str, x: float, y: float, width: float, height: float
    ) -> dict:
        """Annotates the plan with Markdown text.

        :param markdown: text to annotate (in markdown format)
        :type markdown: basestring
        :param x: x-position
        :type x: int
        :param y: y-position
        :type y: int
        :param width: width of the text box
        :type width: int
        :param height: height of the text box
        :type height: int
        :return: None
        :rtype: None
        """
        annotation = {
            "anchor": {"x": width, "y": height},
            "x": x,
            "y": y,
            "markdown": markdown,
        }
        markdown += "\n<!-- annotated by pydent.planner -->"
        self.plan.layout.setdefault("text_boxes", [])
        if self.plan.layout["text_boxes"] is None:
            self.plan.layout["text_boxes"] = []
        if annotation not in self.plan.layout["text_boxes"]:
            self.plan.layout["text_boxes"].append(annotation)
        return annotation

    def annotate_operations(
        self, ops: List[Operation], markdown: str, width: float, height: float
    ) -> dict:
        """Annotates above the operations.

        Estimates the x-midpoint and y and makes a text annotation at
        that location.
        """
        layout = self.layout.ops_to_subgraph(ops)
        return self.annotate_above_layout(markdown, width, height, layout=layout)

    def annotate_above_layout(
        self, markdown: str, width: float, height: float, layout: PlannerLayout = None
    ):
        """Annotates text directly above a layout."""
        if layout is None:
            layout = self.layout

        x = (
            layout.midpoint()[0] + layout.BOX_WIDTH / 2
        )  # adjust so x is midpoint of 'box' on screen
        y = layout.y

        # center the annotation
        x -= width / 2
        y -= height

        # give some space between annotation and operation
        y -= layout.BOX_DELTA_Y / 2

        return self.annotate(markdown, x, y, width, height)

    @property
    def layout(self) -> PlannerLayout:
        return PlannerLayout.from_plan(self.plan)

    @property
    def graph(self):
        return PlannerGraph.from_plan(self.plan)

    def ipython_link(self):
        try:
            from IPython.display import display, HTML

            return display(HTML('<a href="{url}">{url}</a>'.format(url=self.url)))
        except ImportError:
            print("Could not import IPython. This is likely not installed.")

    def roots(self) -> List[FieldValue]:
        """Get field values that have no predecessors (i.e. are 'roots')"""
        roots = []
        for subgraph in get_subgraphs(self._routing_graph()):
            for n in subgraph:
                node = subgraph.nodes[n]
                if len(list(subgraph.predecessors(n))) == 0:
                    root = node["fv"]
                    roots.append(root)
        return roots

    def optimize(self, merge_missing_samples=False) -> "Planner":
        """Optimize the plan, merging wires and operations that have identical
        inputs and outputs.

        .. versionchanged:: 0.1.5a11
            Numerous bug fixes.

        :param merge_missing_samples: if True, the method
            will consider FieldValues with missing inputs
            as identical. By default, this is False, and so
            the method will not merge *any* operation with
            a missing input or output.
        :return: the planner (self)
        """
        optimizer = PlanOptimizer(
            self,
            inherit_logger=self.logger,
            merge_missing_samples=merge_missing_samples,
        )
        return optimizer.optimize()

    def validate(self) -> Dict:
        """Validates sample routes in the plan.

        :return: dictionary of each sample route and validation errors
        :rtype: dict
        """

        def fv_info(fv):
            return "{role} {name} for {ot} {opid}".format(
                role=fv.role,
                name=fv.name,
                ot=fv.operation.operation_type.name,
                opid=fv.operation.id,
            )

        routes = {}
        for subgraph in get_subgraphs(self._routing_graph()):
            values = []
            reasons = []
            root = None
            for n in subgraph:
                node = subgraph.nodes[n]
                fv = node["fv"]
                value = fv.sample
                if value is None and fv.field_type.ftype == "sample":
                    reasons.append("{} has no sample defined".format(fv_info(fv)))
                else:
                    if fv.item:
                        item = fv.item
                        collection = item.as_collection()
                        if collection:
                            item = collection.part(fv.row, fv.column)
                        if item.sample_id != fv.sample.id:
                            reasons.append(
                                "{} has sample_id={} but generated item has sample_"
                                "id={}".format(
                                    fv_info(fv), fv.sample.id, item.sample_id
                                )
                            )
                    values.append(value)
                if len(list(subgraph.predecessors(n))) == 0:
                    root = fv
            if (
                root.item is None
                and root.role == "input"
                and root.field_type.ftype == "sample"
            ):
                reasons.append("{} has no item defined".format(fv_info(fv)))
            values = list(set(values))
            if len(values) > 1:
                reasons.append("different samples defined in route ({})".format(values))
            root_id = "{} for {} ({})".format(
                root.name, root.operation.operation_type.name, self._routing_id(root)
            )
            routes[root_id] = {
                "root_field_value": root,
                "errors": reasons,
                "valid": len(reasons) == 0,
            }
        errors = {k: v for k, v in routes.items() if v["valid"] is not True}
        return errors
        # return routes

    # TODO: implement planner.copy and anonymize the operations and field_values by
    #       removing their ids
    def copy(self) -> "Planner":
        """Return a copy of this planner, with a new anonymous copy of the
        plan.

        Browser cache is copied as well, but model_cache in browser are
        not anonymous
        """
        # copy everything execpt plan, which may be large
        copied = empty_copy(self)
        data = self.__dict__.copy()
        data.pop("_plan")
        copied.__dict__ = deepcopy(data)

        # copy over anonymous copy
        copied._plan = self.plan.copy()

        for op in copied.operations:
            for fv in op.field_values:
                fv.operation = op
        return copied

    def __copy__(self):
        return self.copy()

    @staticmethod
    def combine(plans: List[Plan]) -> "Planner":
        """Merges a list of plans into a single plan by combining operations
        and wires.

        :param plans: list of Aquarium Plans instances
        :return: new Plan
        """
        copied_plans = [c.copy() for c in plans]

        sessions = {p.session for p in plans}
        if len(sessions) > 1:
            raise PlannerException(
                "Cannot combine plans, plans must all derive from same session instance"
            )
        session = sessions.pop()

        new_plan = Planner(session)
        new_plan.plan.operations = []
        for p in copied_plans:
            new_plan.plan.operations += p.plan.operations
            new_plan.plan.wires += p.plan.wires
        return new_plan

    def move_operations(self, ops, to_plan_id, confirm: bool = True):
        """Move a list of operations from this planner to a different plan.

        :param ops: list of operations
        :param to_plan_id: destination plan_id
        :param confirm: whether this required confirmation from the user.
        :return: None
        """
        for op in ops:
            for pa in op.plan_associations:
                pa.plan_id = to_plan_id
                self.session.utils.json_save("PlanAssociation", pa.dump())
        result = confirm(
            "Move {} operations to plan_id = {}? (y|n): ".format(len(ops), to_plan_id)
        )

        if result == "y":
            for op in ops:
                if not self._contains_op(op):
                    raise PlannerException(
                        "Operation {} not contained in this planner".format(op)
                    )
            for op in ops:
                for pa in op.plan_associations:
                    pa.plan_id = to_plan_id
                    self.session.utils.json_save("PlanAssociation", pa.dump())
                    self.plan.operations.remove(op)
            print("moved operations")
        else:
            print("{}: user canceled".format(result))

    def add_from_template(self, plan: Union[int, Plan, "Planner"]) -> List[Operation]:
        """Add Operation and Wires from the provided template plan to this
        planner.

        :param plan: Plan id (int), Plan instance, or Planner instance
        :return: list of operations that were added from template.
        """
        if isinstance(plan, int):
            plan = self.session.Plan.find(plan)
            planner = self.__class__(plan)
        elif isinstance(plan, Plan):
            planner = self.__class__(plan)
        template = planner.copy()
        self.plan.operations += template.operations
        self.plan.wires += template.wires
        return template.operations

    def split(self) -> List["Planner"]:
        """Split the plan into several distinct plans, if possible.

        This first convert the plan  to an operation graph, find
        subgraphs that are not connected to each other,  and create new
        plans based on these subgraphs. This will return anonymous
        copies of all the plans, meaning operations and field_values
        will be anonymized. Sample and items attached to field values
        will remain to avoid re-creating samples and items.
        """

        # copy this plan
        copied_plan = self.copy()

        # get independent operation graphs
        layouts = copied_plan.layout.get_independent_graphs()

        # for each independent graph, make a new plan
        new_plans = []
        for layout in layouts:
            new_plan = Planner(self.session)
            new_plans.append(new_plan)

            # copy over the operations
            opids = list(layout.nxgraph.nodes)
            ops = [layout.nxgraph.nodes[opid]["operation"] for opid in opids]
            wires = []

            # copy over relevant wires
            for wire in copied_plan.plan.wires:
                to_id = wire.destination.operation._primary_key
                from_id = getattr(wire, "source").operation._primary_key
                if to_id in opids or from_id in opids:
                    wires.append(wire)

            new_plan.plan.operations = ops
            new_plan.plan.wires = wires
        return new_plans

    def __add__(self, other: Plan) -> "Planner":
        return self.combine([self, other])

    def __mul__(self, num: int) -> "Planner":
        return self.combine([self] * num)

    def prettify(self) -> None:
        if self.plan.operations:
            self.layout.topo_sort()

    # TODO: implement individual wires and things
    def draw(self):
        layout = self.layout
        edge_colors = []
        for s, e, d in layout.nxgraph.edges(data=True):
            color = "black"
            wire = d["wire"]
            if wire.source.sample is None:
                color = "orange"
            elif wire.destination.sample is None:
                color = "orange"
            elif wire.destination.sample != wire.source.sample:
                color = "red"
            edge_colors.append(color)
        nx.draw(self.layout.nxgraph, edge_colors=edge_colors, pos=self.layout.pos())

    # TODO: autofill implementation
    # def autofill(self):
    # """Fill the unfilled field_values with random items and samples.
    #
    # .. versionadded: 0.1.5a10
    #     Added autofill method.
    #
    # :return:
    # """
    # wired_fv_keys = set()
    # for w in self.plan.wires:
    #     key = w.destination.id or w.destination._primary_key
    #     wired_fv_keys.add(key)
    #
    # needs_sample = []
    # for op in self.operations:
    #     for fv in op.inputs:
    #         key = fv.id or fv._primary_key
    #         if key not in wired_fv_keys:
    #             needs_sample.append(fv)
    #
    # with self.session.with_cache(using_models=True) as sess:
    #     while needs_sample:
    #         for fv in needs_sample:
    #             self.set_field_value(fv, sample=fv.field_type)
    #     skip = []
    #     for fv in needs_items:
    #         afts = fv.field_type.allowable_field_types
    #         aft_tuples = [(aft.sample_type_id, aft.object_type_id) for aft in afts]
    #         samples = sess.Sample.where({'sample_type_id': [aft.sample_type_id for aft in afts]})
    #         items = sess.browser.cached_where({
    #             'sample_id': [s.id for s in samples],
    #             'object_type_id': [aft.object_type_id for aft in afts],
    #         }, 'Item')
    #         items = [item for item in items if
    #                  (item.sample_type_id, item.object_type_id) in aft_tuples]
    #
    #         if not items:
    #             item_query = QueryBuilder.sql({
    #                 'sample_id': [s.id for s in samples],
    #                 'object_type_id': [aft.object_type_id for aft in afts],
    #                 'location': QueryBuilder.Not('deleted')
    #             })
    #             items = self.session.Item.where(item_query)
    #             items = [item for item in items if
    #                      (item.sample_type_id,
    #                       item.object_type_id) in aft_tuples]
    #         if not items:
    #             skip.append(fv)
    #         else:
    #             self.set_field_value_and_propogate(fv, item=items[0])
    #     needs_items = [fv for fv in needs_items if fv.child_item_id is None and fv not in skip]
    #
    # # get unfilled leaf field_values



"""
Exceptions (:mod:`pydent.exceptions`)
=====================================

.. currentmodule:: pydent.exceptions

Trident exceptions
"""
from typing import Type
from warnings import warn


class TridentBaseException(Exception):
    """Base exception for all trident errors."""


class TridentRequestError(TridentBaseException):
    """There was an ambiguous exception that occurred handling your request."""

    def __init__(self, msg, response):
        super().__init__(msg)
        self.response = response


class AquariumError(TridentBaseException):
    """Aquarium raised an error."""


class AquariumModelNotFound(TridentBaseException):
    """Returned when Aquarium could not find a given model."""


class ForbiddenRequestError(TridentBaseException):
    """Raised when Trident attempts to make a request after requests have been
    explicitly turned off."""


class TridentJSONDataIncomplete(TridentBaseException):
    """JSON data contains a null value and may be incomplete."""


class TridentLoginError(TridentBaseException):
    """Trident is not properly connected to the server.

    Verify login credentials are correct.
    """


class TridentTimeoutError(TridentBaseException):
    """Trident took too long to respond."""


class TridentModelNotFoundError(TridentBaseException):
    """Trident could not find model in list of models."""


class AquariumModelError(TridentBaseException):
    """An error occurred with this Aquarium model."""


class NoSessionError(TridentBaseException):
    """There was no session attached to the model, but one is required."""


class SessionAlreadySet(TridentBaseException):
    """Cannot set session to models with a session already set."""


class PlannerException(TridentBaseException):
    """Generic planner Exception."""


class PlannerVerificationException(TridentBaseException):
    """Raised when object is not in plan but is required."""


class TridentDepreciationWarning(DeprecationWarning):
    """Raised when a feature or api is depreciated."""


class AquariumQueryLanguageValidationError(TridentBaseException):
    """Raised when aql is provided with an invalide query."""


# class WarningLimit(object):
#
#     _limits = {}
#
#     def __init__(self, msg: str, limit: int = 1, warning_class: Type = TridentDepreciationWarning):
#         self.msg = msg
#         self.warning_class = warning_class
#         self._limits[msg] = (0, limit)
#
#     def warn(self):
#         a, b = self._limits[self.msg]
#         if a < b:
#             warn(self.warning_class(self.msg))
#             self._limits[self.msg] = (a + 1, b)
#         else:
#             return
#
# def warn_with_limits(msg: str, limit: int):
#     WarningLimit(msg, limit).warn()



"""Marshalling exceptions."""


class MarshallerBaseException(Exception):
    pass


class SchemaRegistryError(MarshallerBaseException):
    """Generic schema registry exception."""


class SchemaException(MarshallerBaseException):
    """A generic schema exception."""


class SchemaModelException(MarshallerBaseException):
    """A generic model exception."""


"""
Field validation exceptions
"""


class FieldValidationError(MarshallerBaseException):
    """A generic field validation error."""


class AllowNoneFieldValidationError(FieldValidationError):
    """A field validation error for getting or setting None values."""


class CallbackValidationError(MarshallerBaseException):
    """A generic callback validation error."""


class RunTimeCallbackAttributeError(AttributeError):
    """Error that occurs during executing a field callback."""


"""
Model validation exceptions
"""


class ModelRegistryError(MarshallerBaseException):
    """Model not found in registry exception."""


class ModelValidationError(MarshallerBaseException):
    """A model validation error."""


class ExceptionCollection(MarshallerBaseException):
    """Context dependent exception for capturing multiple exceptions.

    Call `r` to gather exceptions, upon exiting, a single
    ExceptionCollection will be raised with a summary of all the
    internal exceptions.
    """

    def __init__(self, *args, header=""):
        self.args = args
        self.header = header
        self.errors = None

    def r(self, exception):
        self.errors.append(exception)

    def raise_exception_class(self, exception_class):
        """Raise an exception class, if it was collected."""
        errors = self.group_errors().get(exception_class.__name__, [])
        if errors:
            raise exception_class(errors)

    def group_errors(self):
        grouped = {}
        for e in self.errors:
            grouped.setdefault(e.__class__.__name__, []).append(e)
        return grouped

    def __enter__(self):
        self.errors = []
        return self

    def __exit__(self, *args):
        if self.errors:
            # raise MultipleValidation(self.errors)
            try:
                msg = "{}: {}".format(self.__class__.__name__, self.header)
                group_by_exception = self.group_errors()
                for g, errors in group_by_exception.items():
                    msg += "\n {}(s):".format(g)
                    for i, e in enumerate(errors):
                        msg += "\n  ({}): {}".format(i, e)
                self.args = (msg,)
            except Exception as e:
                raise e.__class__(
                    "{}\nThere was an error raising exceptions {}\n".format(
                        self.errors, e
                    )
                )
            raise self


class MultipleValidationError(ModelValidationError, ExceptionCollection):
    """Model validation exception."""



from copy import deepcopy

import networkx as nx

# where is this method called?
def arr_to_pairs(arr):
    arr1 = arr[:-1]
    arr2 = arr[1:]
    return list(zip(arr1, arr2))


def _id_getter(model):
    return model._primary_key


def to_undirected(graph):
    """.to_undirected is implemented in networkx out of the box, however, it
    suffers from occasional infinite recursion errors during the deepcopy phase
    (unknown as to why)."""
    undirected = nx.Graph()
    copied = deepcopy(graph)
    for n in copied.nodes:
        ndata = copied.nodes[n]
        undirected.add_node(n, **ndata)
    for n1, n2 in copied.edges:
        edata = copied.edges[n1, n2]
        undirected.add_edge(n1, n2, **edata)
    return undirected


def bfs_tree_subgraph(graph, source):
    """Similar to nx.bfs_tree, but nx.bfs_tree does not maintain edge list."""
    nodes = list(nx.dfs_predecessors(graph, source)) + [source]
    return graph.subgraph(nodes)


def get_subgraphs(graph):
    """Get independent subgraphs."""
    node_list = list(graph.nodes)
    subgraphs = []
    while len(node_list) > 0:
        node = node_list[-1]
        subgraph = nx.bfs_tree(to_undirected(graph), node)
        for n in subgraph.nodes:
            node_list.remove(n)
        subgraphs.append(graph.subgraph(subgraph.nodes))
    return subgraphs



import logging
import math
import pprint
import time
import traceback
import weakref
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from itertools import count
from logging import CRITICAL
from logging import DEBUG
from logging import ERROR
from logging import INFO
from logging import WARN
from logging import WARNING
from warnings import warn

from colorlog import ColoredFormatter
from tqdm import tqdm


def utcnow():
    return datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")


def pprint_data(data, width=80, depth=10, max_list_len=20, compact=True, indent=1):
    """Pretty print data."""
    return pprint.pformat(
        condense_long_lists(data, max_list_len=max_list_len),
        indent=indent,
        width=width,
        depth=depth,
        compact=compact,
    )


class LoggableWarning(Warning):
    """Generic logger warning."""


class LoggableException(Exception):
    """Generic exception for loggable class."""


class LoggableHandler(logging.Handler):
    pass


class TqdmLoggingHandler(LoggableHandler):
    def __init__(self, level=logging.NOTSET, tqdm=tqdm):
        super().__init__(level)
        self._tqdm = tqdm
        self.tb_limit = 0

    def emit(self, record):
        try:
            msg = self.format(record)
            self._tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


def condense_long_lists(d, max_list_len=20):
    """Condense the long lists in a dictionary.

    :param d: dictionary to condense
    :type d: dict
    :param max_len: max length of lists to display
    :type max_len: int
    :return:
    :rtype:
    """
    if isinstance(d, dict):
        return_dict = {}
        for k in d:
            return_dict[k] = condense_long_lists(dict(d).pop(k))
        return dict(return_dict)
    elif isinstance(d, list):
        if len(d) > max_list_len:
            g = max_list_len / 2
            return d[: math.floor(g)] + ["..."] + d[-math.ceil(g) :]
        else:
            return d[:]
    return str(d)


class Loggable:
    DEFAULT_FORMAT = "%(log_color)s%(levelname)s - %(name)s - %(asctime)s - %(message)s"
    DEFAULT_COLORS = {
        "DEBUG": "cyan",
        "INFO": "white",
        "SUCCESS:": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    }
    graph = {}
    registered = {}  # id to Loggable
    counter = count()

    def __init__(self, object_or_name, format=None, log_colors=None, tqdm=tqdm):
        """Instantiates a new logger.

        :param object_or_name: The name or object to attach the logger. If o
            bject is provided, the name
            is produced from the object instance.
        :param format: The logger format. Default is found at
            Loggable.DEFAULT_FORMAT
        :param log_colors: The logger colors. Default is found
            Loggable.DEFAULT_COLORS.
        :param tqdm: This tqdm class to instantiate progress bars with.
            Default is tqdm.tqdm.
        """
        if isinstance(object_or_name, str):
            self.name = object_or_name
        else:
            self.name = "{}(id={})".format(
                object_or_name.__class__.__name__, id(object_or_name)
            )

        self.format = format or self.DEFAULT_FORMAT
        self.log_colors = log_colors or self.DEFAULT_COLORS
        self._tqdm = tqdm
        self._id = next(self.counter)

    def _new_logger(self, name, level=logging.ERROR):
        """Instantiate a new logger with the given name.

        If channel handler exists, do not create a new one.
        """
        logger = logging.getLogger(name)
        logger.setLevel(level)
        handlers = self._log_handlers(logger)
        # make stream handler
        if not handlers:
            handler = TqdmLoggingHandler(level, tqdm=self._tqdm)
            handler.tb_limit = 0
            formatter = ColoredFormatter(self.format, log_colors=self.log_colors)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        else:
            handler = handlers[0]
        return logger, handler

    @property
    def _children(self):
        """The logger's children loggers."""
        return self.registered.setdefault(self._id, weakref.WeakValueDictionary())

    @property
    def logger(self):
        """The native logger."""
        return self._new_logger(self.name)[0]

    @property
    def logger_handlers(self):
        """The logging handlers for this logger."""
        return self._log_handlers(self.logger)

    def set_tb_limit(self, limit):
        """Set the throwback limit."""
        for h in self.logger.handlers:
            h.tb_limit = limit
        return self

    def level_name(self):
        """Return the current level name, as a string."""
        return logging._levelToName[self.level()]

    def level(self):
        """Return the current level, as an int."""
        return self.logger_handlers[0].level

    def is_enabled(self, level):
        """Returns whether this logger is enabled for the level specified."""
        level = self._get_level(level)
        return level >= self.level()

    def _get_level(self, level):
        if isinstance(level, str):
            try:
                level = {
                    "INFO": INFO,
                    "ERROR": ERROR,
                    "DEBUG": DEBUG,
                    "CRITICAL": CRITICAL,
                    "WARNING": WARNING,
                    "WARN": WARN,
                }[level.upper()] 
            except KeyError as e:
                raise KeyError("Level {} not recognized.".format(level)) from e
        return level

    def set_level(self, level, tb_limit=None):
        """Sets the level for this logger and its children."""
        level = self._get_level(level)
        self.logger.setLevel(level)
        for h in self.logger.handlers:
            h.setLevel(level)
        if tb_limit is not None:
            self.set_tb_limit(tb_limit)
        for child in self._children.values():
            child.set_level(level, tb_limit=tb_limit)
        return self

    def set_verbose(self, verbose, tb_limit=0):
        """Sets to 'INFO' if True, or 'ERROR' if False."""
        if verbose:
            return self.set_level(logging.INFO, tb_limit)
        else:
            return self.set_level(logging.ERROR, tb_limit)

    pprint_data = staticmethod(pprint_data)

    def tqdm(self, iterable, level, *args, **kwargs):
        """Produce a logged progress bar for an interable."""
        level = self._get_level(level)
        if self.is_enabled(level):
            progress_bar = self._tqdm(iterable, *args, **kwargs)
            progress_bar.set_description(
                "{:8} {}".format(logging._levelToName[level], progress_bar.desc)
            )
            return progress_bar
        else:
            return iterable

    def _add_child(self, other):
        self._children[other._id] = other
        return other

    def _log_handlers(self, logger):
        return [h for h in logger.handlers if issubclass(type(h), LoggableHandler)]

    def log(self, msg, level):
        """Log at specified level."""
        level = self._get_level(level)
        logger = self.logger
        logger.log(level, msg)
        if logger.isEnabledFor(level):
            tb_limit = self.logger_handlers[0].tb_limit
            if tb_limit:
                traceback.print_stack(limit=tb_limit)
        return self

    def critical(self, msg):
        """Log critical error."""
        return self.log(msg, CRITICAL)

    def error(self, msg):
        """Log error."""
        return self.log(msg, ERROR)

    def warn(self, msg):
        """Log warning."""
        return self.log(msg, WARNING)

    def info(self, msg):
        """Log info."""
        return self.log(msg, INFO)

    def debug(self, msg):
        """Log debug."""
        return self.log(msg, DEBUG)

    def __copy__(self):
        return self.copy()

    def copy(self, name=None):
        """Copy this logger."""
        copied = self.__class__(
            name or self.name,
            format=self.format,
            log_colors=self.log_colors,
            tqdm=self._tqdm,
        )
        copied.set_level(self.level())
        return copied

    def spawn(self, name=None):
        """Spawn a child logger.

        Child loggers will automatically inherit the level of their
        parent. This inheritance occurs after creation as well.
        """
        return self._add_child(self.copy(name))

    def timeit(self, level, prefix=""):
        """Spawn a logger that computes the time it takes to run commands.

        The start time and total time will be logged.
        """
        child = TimedLoggable(
            self.name,
            level,
            prefix=prefix,
            format=self.format,
            log_colors=self.log_colors,
            tqdm=self._tqdm,
        )
        child.set_level(self.level())
        return self._add_child(child)

    def track(self, level, total=None, desc=None):
        """Spawn a progress bar logger, which can update a progress bar."""
        child = ProgressLoggable(
            self.name,
            level,
            desc=desc,
            total=total,
            format=self.format,
            log_colors=self.log_colors,
            tqdm=self._tqdm,
        )
        child.set_level(self.level())
        return self._add_child(child)

    # alias for 'track', which is hard to remember
    pbar = track

    def __call__(self, name):
        return self.spawn(name)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    # def __str__(self):
    #     return "<{} {}>".format(self.__class__.__name__, self.logger)
    #
    # def __repr__(self):
    #     return str(self)


class RenamedLoggable(Loggable):
    def __init__(self, object_or_name, format=None, log_colors=None, tqdm=tqdm):
        new_name = "{}({})".format(self.__class__.__name__, object_or_name)
        super().__init__(new_name, format, log_colors, tqdm)


class LockedLoggable(RenamedLoggable):
    def __init__(self, object_or_name, level, format=None, log_colors=None, tqdm=tqdm):
        super().__init__(object_or_name, format, log_colors, tqdm)
        self.locked_level = level

    def is_enabled(self, level=None):
        level = level or self.locked_level
        return super().is_enabled(level)

    def log(self, msg, level=None):
        level = level or self.locked_level
        super().log(msg, level)


class Enterable(ABC):
    @abstractmethod
    def enter(self):
        pass

    @abstractmethod
    def exit(self):
        pass

    def __enter__(self):
        return self.enter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.exit()
        except Exception as e:
            warn(
                LoggableWarning(
                    "Exception during 'exit' of {}: {}\n{}".format(self, type(e), e)
                )
            )


class ProgressLoggable(Enterable, LockedLoggable):
    def __init__(
        self,
        object_or_name,
        level,
        total,
        desc,
        format=None,
        log_colors=None,
        tqdm=tqdm,
    ):
        super().__init__(
            object_or_name, level, format=format, log_colors=log_colors, tqdm=tqdm
        )
        self.desc = desc
        self.total = total
        self.pbar = None

    def update(self, x=1, msg=""):
        if self.pbar is None:
            raise LoggableException(
                "Progress bar not instantiated. Run 'enter()' or 'with' to instantiate."
            )
        if msg:
            self.log(msg)
        if self.is_enabled():
            self.pbar.update(x)

    def __call__(self, iterable, *args, **kwargs):
        d = {"desc": self.desc}
        d.update(kwargs)
        return self.tqdm(iterable, self.locked_level, *args, **d)

    def enter(self):
        if self.is_enabled():
            self.pbar = tqdm(total=self.total)
        else:
            self.pbar = 0
        return self

    def exit(self):
        if self.is_enabled():
            self.pbar.close()
        else:
            self.pbar = 0
        return self


class TimedLoggable(Enterable, LockedLoggable):
    def __init__(
        self, object_or_name, level, prefix="", format=None, log_colors=None, tqdm=tqdm
    ):
        super().__init__(
            object_or_name, level, format=format, log_colors=log_colors, tqdm=tqdm
        )
        now = time.time()
        self.t1 = now
        self.t2 = None
        self.time = None
        self.prefix = prefix

    def log(self, msg, level=None):
        if self.prefix:
            msg = '{}("{}"): {}'.format(self.__class__.__name__, self.prefix, msg)
        super().log(msg, level)

    def enter(self):
        now = time.time()
        self.t1 = now
        self.log("Started at {}".format(self.t1))
        return self

    def exit(self):
        t2 = time.time()
        self.t2 = t2
        self.time = self.t2 - self.t1
        self.log("Finished in {}.".format(self.time))
        return self


class LoggableFactory:
    def __init__(self, format=None, log_colors=None, tqdm=tqdm):
        self.format = format
        self.log_colors = log_colors
        self._tqdm = tqdm

    def __call__(self, name):
        return Loggable(
            name, format=self.format, log_colors=self.log_colors, tqdm=self._tqdm
        )



from collections.abc import Mapping
from collections.abc import Sequence
from typing import Callable
from typing import Generator
from typing import Generic
from typing import Iterable
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union


IndexType = Union[Tuple[Union[int, slice], Tuple[Union[int, slice]], int, slice]]
T = TypeVar("T")


class MatrixMapping(Sequence, Generic[T]):
    """A 2D matrix that behaves like a numpy matrix more user friendly setters
    and getters."""

    def __init__(
        self,
        data: Iterable[Iterable[T]],
        setter: Callable = None,
        getter: Callable = None,
    ):
        self.data = data
        self.validate()
        self.setter = setter
        if getter is None:
            self.getter = lambda x: x
        else:
            self.getter = getter

    @property
    def dimensions(self) -> Tuple[int, int]:
        return len(self.data), len(self.data[0])

    def set_item(self, r: int, c: int, value: T):
        if self.setter is None:
            self.data[r][c] = value
        else:
            return self.setter(self.data, r, c, value)

    @staticmethod
    def _flatten(arr):
        for a in arr:
            if isinstance(a, list):
                for _a in a:
                    yield _a
            else:
                yield a

    def validate(self):
        if not self._is_rect():
            raise ValueError("Data rows have different lengths")

    def _is_rect(self):
        return len({len(d) for d in self.data}) == 1

    @property
    def _index_matrix(self):
        index_matrix = []
        for r in range(len(self.data)):
            col_indices = list(range(len(self.data[0])))
            tuples = [(r, c) for c in col_indices]
            index_matrix.append(tuples)
        return MatrixMapping(index_matrix)

    @property
    def _index_view(self):
        return self._index_matrix

    def _iter_indices(self, index: IndexType):
        indices = self._index_view[index]

        if isinstance(indices, tuple):
            indices = [[indices]]

        for rc in self._flatten(indices):
            if isinstance(rc, list):
                for r, c in rc:
                    yield r, c
            else:
                yield rc

    def iter_flatten(self) -> Generator[T, None, None]:
        for row in self.data:
            for col in row:
                yield col

    def flatten(self) -> List[T]:
        return list(self.iter_flatten())

    def __len__(self) -> int:
        return len(self.data)

    def __contains__(self, value: T) -> bool:
        for a in self.flatten():
            if a == value:
                return True
        return False

    def __iter__(self) -> Generator[T, None, None]:
        return self.data.__iter__()

    def __setitem__(self, index: IndexType, value: T):
        for r, c in self._iter_indices(index):
            self.set_item(r, c, value)

    #     def __delitem__(self, index: IndexType):
    #         for r, c in self._iter_indices(index):
    #             print(r, c)
    #             print(self.data)
    #             del self.data[r][c]

    def __getitem__(self, index: IndexType) -> Union[T, Iterable[T]]:
        if isinstance(index, slice):
            rows = self.data[index]
            return [[self.getter(d) for d in row] for row in rows]
        elif isinstance(index, tuple):
            if isinstance(index[0], slice):
                rows = self.data[index[0]]
                if isinstance(index[1], slice):
                    return [[self.getter(d) for d in row[index[1]]] for row in rows]
                return [self.getter(row[index[1]]) for row in rows]
            else:
                row = self.data[index[0]]
                if isinstance(index[1], slice):
                    return [self.getter(d) for d in row[index[1]]]
                return self.getter(row[index[1]])
        return [self.getter(d) for d in self.data[index]]

    def get(self, index: IndexType, as_mapping=False):
        if as_mapping:
            return self.__class__(self[index])
        else:
            return self[index]

    def __eq__(self, other):
        return self[:] == other[:]

    def __str__(self):
        return str(self[:])


class MatrixMappingFactory(Mapping):
    """Different ways to view the same set of data."""

    def __init__(self, data):
        self.data = data
        self.views = {}

    def new(self, name, setter: Callable = None, getter: Callable = None):
        self.views[name] = MatrixMapping(self.data, setter=setter, getter=getter)

    def __getitem__(self, key):
        return self.views[key]

    def __iter__(self):
        return self.views.__iter__()

    def __len__(self):
        return len(self.views)


# def set_part(data, r, c, value):
#     if data[r][c] is None:
#         print("NEW!")
#     if value is None:
#         print("deleting")
#
#
# from pydent.models import Collection
#
#
# def _empty(self):
#     nrows, ncols = self.dimensions
#     data = []
#     for r in range(nrows):
#         data.append([None] * ncols)
#     return data
#
#
# @property
# def part_association_matrix(self):
#     data = self._empty()
#     for assoc in self.part_associations:
#         data[assoc.row][assoc.column] = assoc
#     return data
#
#
# @property
# def part_matrix(self):
#     data = self._empty()
#     for r, row in enumerate(self.part_association_matrix):
#         for c, assoc in enumerate(row):
#             if assoc is not None:
#                 data[r][c] = assoc.part
#     return data
#
#
# @property
# def data_association_matrix(self):
#     data = self._empty()
#     for r, row in enumerate(self.part_matrix):
#         for c, assoc in enumerate(row):
#             if assoc is not None:
#                 data[r][c] = assoc.part
#     return data
#
#
# @property
# def matrix(self):
#     data = self._empty()
#     for r, row in enumerate(self.part_association_matrix):
#         for c, assoc in enumerate(row):
#             if assoc is not None:
#                 if assoc.part.sample_id:
#                     data[r][c] = assoc.part.sample_id
#                 else:
#                     data[r][c] = assoc.part.sample.id
#     return data
#
#
# @property
# def part_matrix_view(self):
#     return MatrixMapping(self.part_matrix)
#
#
# @property
# def matrix_view(self):
#     return MatrixMapping(self.matrix, setter=self._set_part)
#
#
# @property
# def part_association_matrix_view(self):
#     return MatrixMapping(self.part_association_matrix)
#
#
# def __getitem__(self, index: IndexType):
#     return self.matrix_view[index]
#
#
# def __setitem__(self, index: IndexType, value: Union[int, Sample]):
#     matrix_view = self.matrix_view[index]
#     matrix_view[index] = value
#
#
# def _set_part(self, data: List[List[int]], r: int, c: int, value: Union[int, Sample]):
#     pass
#
#
# # TODO: versionadded
# # TODO: versionchanged
# # TODO: documentation
# # TODO: better way to
# # TODO: data_association viewer
#
# Collection._empty = _empty
# Collection.part_association_matrix = part_association_matrix
# Collection.part_matrix = part_matrix
# Collection.matrix = matrix
#
# Collection.part_association_matrix_view = part_association_matrix_view
# Collection.part_matrix_view = part_matrix_view
# Collection.matrix_view = matrix_view
#
# Collection.__getitem__ = __getitem__
# Collection.__setitem__ = __setitem__
# Collection._set_part = _set_part
#
# factory = MatrixMappingFactory(collection.part_association_matrix)
#
#
# def part_getter(association):
#     if association is not None:
#         return association.part
#
#
# def value_getter(association):
#     if association is not None:
#         return association.part.sample_id or association.part.sample.id
#
#
# def data_getter(association):
#     if association is not None:
#         return association.part.data_associations
#
#
# factory.new('part', getter=part_getter)
# factory.new('matrix', getter=value_getter)
# factory.new('data', getter=data_getter)
# factory['data'][:4, :3]
#
# # del matrix[:1]
# # matrix.data
# #     def __setitem__(self, index: IndexType, value: Any):
# #         if isinstance(index, int):
# #             for i in range(self.data[index])
#
#
# # def part_association_view(self, index):
# #     associations = []
# #     for row in self.dimensions[0]:
# #         assoc_row = [None] * self.dimensions[1]
# #         for column in self.dimensions[1]:
#
#
# # def __getitem__(self, index):
# #     print(index)
# #     if isinstance(index, int) or isinstance(index, slice):
# #         return self.matrix[index]
# #     elif isinstance(index, tuple):
# #         a = self.matrix[index[0]]
# #         return [_a[index[1]] for _a in a]
# #     else:
# #         raise IndexError("Type '{}' cannot be used as an index".format(type(index)))
#
#
# # def __setitem__(self,
# #                 index: Union[Tuple[Union[int, slice], ...], int, slice],
# #                 value: Union[int, Sample]):
# #     if isinstance(index, int):
# #         p
#
#
# # from pydent.models import Collection
# # Collection.__getitem__ = __getitem__
#
# # collection[0]



"""
Models (:mod:`pydent.models`)
=============================

.. _models:

.. currentmodule:: pydent.models

This module contains classes for various Aquarium objects.

Models are loaded and initialized from a :class:`AqSession <pydent.aqsession.AqSession>`
instance.

.. code-block:: python

    sample = session.Sample.one()
    last50 = session.Item.last(50)
    first10 = session.SampleType.first(10)
    mysamples = session.Sample.last(10, query={'user_id': 66})

.. note::
    If using an interactive developer, type `session.` and hit `[TAB]` to
    view possible models. Most models have standard method interfaces,
    such as `find`, `where`, `one`, `first`, `last`, `all`.

From a model instance, attributes can be accessed directly from the model.

.. code-block:: python

    print(sample.name)
    print(sample.id)
    # "LJk12345"
    # 3

Calling some attributes will automatically make a new *request*. For example,
calling `sample_type` from `sample` will initialize a request for information
about that sample's `sample_type`:

.. code-block:: python

    print(sample.sample_type.name)
    # "Plasmid Stock"

If we call `sample_type` again, a new request *will not be made*, rather
the model will return the previously returned data. If a new request
is desired, we can reset the field using the
:meth:`reset_field <pydent.base.ModelBase.reset_field>` method:

.. code-block:: python

    assert sample.is_deserialized('sample_type')
    sample.reset_field('sample_type')
    assert not sample.is_deserialized('sample_type')


    # now calling 'sample_type' again will initialize a new request
    # which will be cached. This can be checked using `is_deserialized`
    sample.sample_type
    assert sample.is_deserialized('sample_type')

.. versionadded:: 0.1.5a7
    :meth:`reset_field <pydent.base.ModelBase.reset_field>` method added to reset
    fields.

.. versionadded:: 0.1.5a7
    :meth:`is_deserialized <pydent.base.ModelBase.is_deserialized>` can be called
    to check if a relation has already been deserialized

Models can be dumped to their JSON data or loaded from data

.. code-block:: python

    # dump the data
    data = sample.dump()

    # load from the data
    sample2 = session.Sample.load(data)

.. warning::
    Most model methods require an attached :class:`Session <pydent.aqsession.AqSession>`
    instance. Importing models from *pydent.models* will result in orphaned models.
    Preferrably, if you need to initialize or load new models, use
    `session.<ModelName>(*args, **kwargs)`, replacing <ModelName> with the name
    of the model class. Alternatively, you can attach a Session instance using
    :meth:`model.connect_to_session(session) <pydent.base.ModelBase.connect_to_session>`

Model Classes
-------------

Operations
^^^^^^^^^^

.. autosummary::
    :toctree: generated/

    Operation
    OperationType
    Code
    Library
    Job
    JobAssociation

Plans
^^^^^

.. autosummary::
    :toctree: generated/

    Plan
    PlanAssociation
    Wire

Inventory
^^^^^^^^^

.. autosummary::
    :toctree: generated/

    Item
    Collection
    ObjectType
    Sample
    SampleType
    PartAssociation

Data and IO
^^^^^^^^^^^

.. autosummary::
    :toctree: generated/

    DataAssociation
    FieldValue
    FieldType
    AllowableFieldType

User
^^^^

.. autosummary::
    :toctree: generated/

    Account
    Budget
    Group
    Invoice
    Membership
    User
    UserBudgetAssociation

Model Mixins
------------

.. currentmodule:: pydent.models

These classes provide some models with common methods such as saving or deleting
models from the server or associating data.

.. autosummary::
    :toctree: generated/

    crud_mixin.CreateMixin
    crud_mixin.DeleteMixin
    crud_mixin.JSONDeleteMixin
    crud_mixin.JSONSaveMixin
    crud_mixin.SaveMixin
    crud_mixin.UpdateMixin
    data_associations.DataAssociatorMixin
    field_value_mixins.FieldMixin
    field_value_mixins.FieldTypeInterface
    field_value_mixins.FieldValueInterface


Field Relationships
-------------------

.. currentmodule:: pydent.relationships

Field relationships are attribute descriptors that may call
information from the Aquarium server if necesssary.

.. autosummary::
    :toctree: generated/

    BaseRelationship
    BaseRelationshipAccessor
    Function
    HasMany
    HasManyGeneric
    HasManyThrough
    HasOne
    HasOneFromMany
    JSON
    Many
    One
    Raw

Exceptions
^^^^^^^^^^

.. autosummary::
    :toctree: generated/

    FieldValidationError

Utilities
---------

.. currentmodule:: pydent.marshaller

.. autosummary::
    :toctree: generated/

    add_schema
"""
from pydent.models.code import Code
from pydent.models.code import Library
from pydent.models.data_associations import DataAssociation
from pydent.models.data_associations import Upload
from pydent.models.field_value import AllowableFieldType
from pydent.models.field_value import FieldType
from pydent.models.field_value import FieldValue
from pydent.models.inventory import Collection
from pydent.models.inventory import Item
from pydent.models.inventory import ObjectType
from pydent.models.inventory import PartAssociation
from pydent.models.job import Job
from pydent.models.job import JobAssociation
from pydent.models.operation import Operation
from pydent.models.operation import OperationType
from pydent.models.plan import Plan
from pydent.models.plan import PlanAssociation
from pydent.models.plan import Wire
from pydent.models.sample import Sample
from pydent.models.sample import SampleType
from pydent.models.user import Account
from pydent.models.user import Budget
from pydent.models.user import Group
from pydent.models.user import Invoice
from pydent.models.user import Membership
from pydent.models.user import User
from pydent.models.user import UserBudgetAssociation

__all__ = [
    "Account",
    "AllowableFieldType",
    "Budget",
    "Code",
    "Collection",
    "DataAssociation",
    "FieldType",
    "FieldValue",
    "Group",
    "Invoice",
    "Item",
    "Job",
    "JobAssociation",
    "Library",
    "Membership",
    "ObjectType",
    "Operation",
    "OperationType",
    "PartAssociation",
    "Plan",
    "PlanAssociation",
    "Sample",
    "SampleType",
    "Upload",
    "User",
    "UserBudgetAssociation",
    "Wire",
]
